# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM bindings **************************************************************

import atexit
import ctypes
import enum
import functools
import inspect
from llvmlite import ir
import numpy as np
import os
import re
from typing import Set
import weakref
from psyneulink.core.scheduling.time import Time
from psyneulink.core.globals.sampleiterator import SampleIterator
from psyneulink.core.globals.utilities import ContentAddressableList
from psyneulink.core import llvm as pnlvm
from . import codegen
from .debug import debug_env

__all__ = ['LLVMBuilderContext', '_modules', '_find_llvm_function']


_modules: Set[ir.Module] = set()
_all_modules: Set[ir.Module] = set()
_struct_count = 0


@atexit.register
def module_count():
    if "stat" in debug_env:
        print("Total LLVM modules: ", len(_all_modules))
        print("Total structures generated: ", _struct_count)
        s = LLVMBuilderContext.get_global()
        print("Total generations by global context: {}".format(s._llvm_generation))
        print("Object cache in global context: {} hits, {} misses".format(s._stats["cache_requests"] - s._stats["cache_misses"], s._stats["cache_misses"]))
        for stat in ("input", "output", "param", "state", "data"):
            gen_stat = s._stats[stat + "_structs_generated"]
            print("Total {} structs generated by global context: {}".format(stat, gen_stat))
        print("Total python types converted by global context: {}".format(s._stats["types_converted"]))




_BUILTIN_PREFIX = "__pnl_builtin_"
_builtin_intrinsics = frozenset(('pow', 'log', 'exp'))


class _node_wrapper():
    def __init__(self, composition, node):
        self._comp = composition
        self._node = node

    def _gen_llvm_function(self, *, ctx, tags:frozenset):
        return codegen.gen_node_wrapper(ctx, self._comp, self._node, tags=tags)

def _comp_cached(func):
    @functools.wraps(func)
    def wrapper(bctx, obj):
        try:
            obj_cache = bctx._cache.setdefault(obj, dict())
        except TypeError:  # 'super()' references can't be cached
            obj_cache = None
        else:
            if func in obj_cache:
                return obj_cache[func]
        val = func(bctx, obj)
        if obj_cache is not None:
            obj_cache[func] = val
        return val

    return wrapper


class LLVMBuilderContext:
    __global_context = None
    __uniq_counter = 0
    _llvm_generation = 0
    int32_ty = ir.IntType(32)
    float_ty = ir.DoubleType()
    bool_ty = ir.IntType(1)

    def __init__(self):
        self._modules = []
        self._cache = weakref.WeakKeyDictionary()
        self._stats = { "cache_misses":0,
                        "cache_requests":0,
                        "types_converted":0,
                        "param_structs_generated":0,
                        "state_structs_generated":0,
                        "data_structs_generated":0,
                        "input_structs_generated":0,
                        "output_structs_generated":0,
                      }

    def __enter__(self):
        module = ir.Module(name="PsyNeuLinkModule-" + str(LLVMBuilderContext._llvm_generation))
        self._modules.append(module)
        LLVMBuilderContext._llvm_generation += 1
        return self

    def __exit__(self, e_type, e_value, e_traceback):
        assert len(self._modules) > 0
        module = self._modules.pop()
        _modules.add(module)
        _all_modules.add(module)

    @property
    def module(self):
        assert len(self._modules) > 0
        return self._modules[-1]

    @classmethod
    def get_global(cls):
        if cls.__global_context is None:
            cls.__global_context = LLVMBuilderContext()
        return cls.__global_context

    @classmethod
    def get_unique_name(cls, name: str):
        cls.__uniq_counter += 1
        name = re.sub(r"[^a-zA-Z0-9_]", "_", name)
        return name + '_' + str(cls.__uniq_counter)

    def get_builtin(self, name: str, args=[], function_type=None):
        if name in _builtin_intrinsics:
            return self.import_llvm_function(_BUILTIN_PREFIX + name)
        if name in ('maxnum'):
            function_type = pnlvm.ir.FunctionType(args[0], [args[0], args[0]])
        return self.module.declare_intrinsic("llvm." + name, args, function_type)

    def create_llvm_function(self, args, component, name=None, *, return_type=ir.VoidType(), tags:frozenset=frozenset()):
        name = "_".join((str(component), *tags)) if name is None else name

        # Builtins are already unique and need to keep their special name
        func_name = name if name.startswith(_BUILTIN_PREFIX) else self.get_unique_name(name)
        func_ty = pnlvm.ir.FunctionType(return_type, args)
        llvm_func = pnlvm.ir.Function(self.module, func_ty, name=func_name)
        llvm_func.attributes.add('argmemonly')
        for a in llvm_func.args:
            if isinstance(a.type, ir.PointerType):
                a.attributes.add('nonnull')

        metadata = self.get_debug_location(llvm_func, component)
        if metadata is not None:
            scope = dict(metadata.operands)["scope"]
            llvm_func.set_metadata("dbg", scope)

        # Create entry block
        block = llvm_func.append_basic_block(name="entry")
        builder = pnlvm.ir.IRBuilder(block)
        builder.debug_metadata = metadata

        return builder

    def gen_llvm_function(self, obj, *, tags:frozenset) -> ir.Function:
        obj_cache = self._cache.setdefault(obj, dict())

        self._stats["cache_requests"] += 1
        if tags not in obj_cache:
            self._stats["cache_misses"] += 1
            with self:
                obj_cache[tags] = obj._gen_llvm_function(ctx=self, tags=tags)
        return obj_cache[tags]

    def import_llvm_function(self, fun, *, tags:frozenset=frozenset()) -> ir.Function:
        """
        Get function handle if function exists in current modele.
        Create function declaration if it exists in a older module.
        """
        if isinstance(fun, str):
            f = _find_llvm_function(fun, _all_modules | {self.module})
        else:
            f = self.gen_llvm_function(fun, tags=tags)

        # Add declaration to the current module
        if f.name not in self.module.globals:
            decl_f = ir.Function(self.module, f.type.pointee, f.name)
            assert decl_f.is_declaration
            return decl_f
        return f

    @staticmethod
    def get_debug_location(func: ir.Function, component):
        if "debug_info" not in debug_env:
            return

        mod = func.module
        path = inspect.getfile(component.__class__) if component is not None else "<pnl_builtin>"
        d_version = mod.add_metadata([ir.IntType(32)(2), "Dwarf Version", ir.IntType(32)(4)])
        di_version = mod.add_metadata([ir.IntType(32)(2), "Debug Info Version", ir.IntType(32)(3)])
        flags = mod.add_named_metadata("llvm.module.flags")
        if len(flags.operands) == 0:
            flags.add(d_version)
            flags.add(di_version)
        cu = mod.add_named_metadata("llvm.dbg.cu")
        di_file = mod.add_debug_info("DIFile", {
            "filename": os.path.basename(path),
            "directory": os.path.dirname(path),
        })
        di_func_type = mod.add_debug_info("DISubroutineType", {
            # None as `null`
            "types": mod.add_metadata([None]),
        })
        di_compileunit = mod.add_debug_info("DICompileUnit", {
            "language": ir.DIToken("DW_LANG_Python"),
            "file": di_file,
            "producer": "PsyNeuLink",
            "runtimeVersion": 0,
            "isOptimized": False,
        }, is_distinct=True)
        cu.add(di_compileunit)
        di_func = mod.add_debug_info("DISubprogram", {
            "name": func.name,
            "file": di_file,
            "line": 0,
            "type": di_func_type,
            "isLocal": False,
            "unit": di_compileunit,
        }, is_distinct=True)
        di_loc = mod.add_debug_info("DILocation", {
            "line": 0, "column": 0, "scope": di_func,
        })
        return di_loc

    @_comp_cached
    def get_input_struct_type(self, component):
        self._stats["input_structs_generated"] += 1
        if hasattr(component, '_get_input_struct_type'):
            return component._get_input_struct_type(self)

        default_var = component.defaults.variable
        return self.convert_python_struct_to_llvm_ir(default_var)

    @_comp_cached
    def get_output_struct_type(self, component):
        self._stats["output_structs_generated"] += 1
        if hasattr(component, '_get_output_struct_type'):
            return component._get_output_struct_type(self)

        default_val = component.defaults.value
        return self.convert_python_struct_to_llvm_ir(default_val)

    @_comp_cached
    def get_param_struct_type(self, component):
        self._stats["param_structs_generated"] += 1
        if hasattr(component, '_get_param_struct_type'):
            return component._get_param_struct_type(self)

        def _param_struct(p):
            val = p.get(None)   # this should use defaults
            if hasattr(val, "_get_compilation_params") or \
               hasattr(val, "_get_param_struct_type"):
                return self.get_param_struct_type(val)
            if isinstance(val, ContentAddressableList):
                return ir.LiteralStructType(self.get_param_struct_type(x) for x in val)
            elif p.name == 'matrix':   # Flatten matrix
                val = np.asfarray(val).flatten()
            elif p.name == 'num_estimates':  # Should always be int
                val = np.int32(0) if val is None else np.int32(val)
            elif np.ndim(val) == 0 and component._is_param_modulated(p):
                val = [val]   # modulation adds array wrap
            return self.convert_python_struct_to_llvm_ir(val)

        elements = map(_param_struct, component._get_compilation_params())
        return ir.LiteralStructType(elements)

    @_comp_cached
    def get_state_struct_type(self, component):
        self._stats["state_structs_generated"] += 1
        if hasattr(component, '_get_state_struct_type'):
            return component._get_state_struct_type(self)

        def _state_struct(p):
            val = p.get(None)   # this should use defaults
            if hasattr(val, "_get_compilation_state") or \
               hasattr(val, "_get_state_struct_type"):
                return self.get_state_struct_type(val)
            if isinstance(val, ContentAddressableList):
                return ir.LiteralStructType(self.get_state_struct_type(x) for x in val)
            struct = self.convert_python_struct_to_llvm_ir(val)
            return ir.ArrayType(struct, p.history_min_length + 1)

        elements = map(_state_struct, component._get_compilation_state())
        return ir.LiteralStructType(elements)

    @_comp_cached
    def get_data_struct_type(self, component):
        self._stats["data_structs_generated"] += 1
        if hasattr(component, '_get_data_struct_type'):
            return component._get_data_struct_type(self)

        return ir.LiteralStructType([])

    def get_node_wrapper(self, composition, node):
        cache = getattr(composition, '_node_wrappers', None)
        if cache is None:
            cache = dict()
            setattr(composition, '_node_wrappers', cache)
        return cache.setdefault(node, _node_wrapper(composition, node))

    def convert_python_struct_to_llvm_ir(self, t):
        self._stats["types_converted"] += 1
        if t is None:
            return ir.LiteralStructType([])
        elif type(t) is list:
            if len(t) == 0:
                return ir.LiteralStructType([])
            elems_t = [self.convert_python_struct_to_llvm_ir(x) for x in t]
            if all(x == elems_t[0] for x in elems_t):
                return ir.ArrayType(elems_t[0], len(elems_t))
            return ir.LiteralStructType(elems_t)
        elif type(t) is tuple:
            elems_t = [self.convert_python_struct_to_llvm_ir(x) for x in t]
            if len(elems_t) > 0 and all(x == elems_t[0] for x in elems_t):
                return ir.ArrayType(elems_t[0], len(elems_t))
            return ir.LiteralStructType(elems_t)
        elif isinstance(t, enum.Enum):
            # FIXME: Consider enums of non-int type
            assert all(round(x.value) == x.value for x in type(t))
            return self.int32_ty
        elif isinstance(t, (int, float, np.floating)):
            return self.float_ty
        elif isinstance(t, np.integer):
            # Python 'int' is handled above as it is the default type for '0'
            return ir.IntType(t.nbytes * 8)
        elif isinstance(t, np.ndarray):
            return self.convert_python_struct_to_llvm_ir(t.tolist())
        elif isinstance(t, np.random.RandomState):
            return pnlvm.builtins.get_mersenne_twister_state_struct(self)
        elif isinstance(t, Time):
            return ir.ArrayType(self.int32_ty, len(Time._time_scale_attr_map))
        elif isinstance(t, SampleIterator):
            if isinstance(t.generator, list):
                return ir.ArrayType(self.float_ty, len(t.generator))
            # Generic iterator is {start, increment, count}
            return ir.LiteralStructType((self.float_ty, self.float_ty, self.int32_ty))
        assert False, "Don't know how to convert {}".format(type(t))


def _find_llvm_function(name: str, mods=_all_modules) -> ir.Function:
    f = None
    for m in mods:
        if name in m.globals:
            f = m.get_global(name)

    if not isinstance(f, ir.Function):
        raise ValueError("No such function: {}".format(name))
    return f


def _gen_cuda_kernel_wrapper_module(function):
    module = ir.Module(name="wrapper_" + function.name)

    decl_f = ir.Function(module, function.type.pointee, function.name)
    assert decl_f.is_declaration

    wrapper_type = ir.FunctionType(ir.VoidType(), (*function.type.pointee.args,
                                                   ir.IntType(32)))
    kernel_func = ir.Function(module, wrapper_type, function.name + "_cuda_kernel")
    block = kernel_func.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)

    # Calculate global id of a thread in x dimension
    intrin_ty = ir.FunctionType(ir.IntType(32), [])
    tid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.tid.x")
    ntid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ntid.x")
    ctaid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ctaid.x")
    global_id = builder.mul(builder.call(ctaid_x_f, []), builder.call(ntid_x_f, []))
    global_id = builder.add(global_id, builder.call(tid_x_f, []))

    # Check global id and exit if we're over
    should_quit = builder.icmp_unsigned(">=", global_id, kernel_func.args[-1])
    with builder.if_then(should_quit):
        builder.ret_void()

    # Index all pointer arguments. Ignore the thread count argument
    args = list(kernel_func.args)[:-1]
    indexed_args = []

    # There are 6 arguments to evaluate:
    # comp_param, comp_state, allocations, output, input, comp_data
    is_grid_evaluate = len(args) == 6

    # Runs need special handling. data_in and data_out are one dimensional,
    # but hold entries for all parallel invocations.
    is_comp_run = len(args) == 7
    if is_comp_run:
        runs_count = args[5]
        input_count = args[6]

    for i, arg in enumerate(args):
        # Don't adjust #inputs and #trials
        if isinstance(arg.type, ir.PointerType):
            offset = global_id
            if is_comp_run:
                # #runs and #trials needs to be the same
                if i >= 5:
                    offset = ir.IntType(32)(0)
                # data arrays need special handling
                elif is_comp_run and i == 4:  # data_out
                    offset = builder.mul(global_id, builder.load(runs_count))
                elif is_comp_run and i == 3:  # data_in
                    offset = builder.mul(global_id, builder.load(input_count))
            elif is_grid_evaluate:
                # all but #2 and #3 are shared
                if i != 2 and i != 3:
                    offset = ir.IntType(32)(0)

            arg = builder.gep(arg, [offset])

        indexed_args.append(arg)
    builder.call(decl_f, indexed_args)
    builder.ret_void()

    # Add kernel mark metadata
    module.add_named_metadata("nvvm.annotations", [kernel_func, "kernel", ir.IntType(32)(1)])

    return module


@functools.lru_cache(maxsize=128)
def _convert_llvm_ir_to_ctype(t: ir.Type):
    type_t = type(t)

    if type_t is ir.VoidType:
        return None
    elif type_t is ir.IntType:
        if t.width == 8:
            return ctypes.c_int8
        elif t.width == 16:
            return ctypes.c_int16
        elif t.width == 32:
            return ctypes.c_int32
        elif t.width == 64:
            return ctypes.c_int64
        else:
            assert False, "Unknown integer type: {}".format(type_t)
    elif type_t is ir.DoubleType:
        return ctypes.c_double
    elif type_t is ir.FloatType:
        return ctypes.c_float
    elif type_t is ir.PointerType:
        pointee = _convert_llvm_ir_to_ctype(t.pointee)
        ret_t = ctypes.POINTER(pointee)
    elif type_t is ir.ArrayType:
        element_type = _convert_llvm_ir_to_ctype(t.element)
        ret_t = element_type * len(t)
    elif type_t is ir.LiteralStructType:
        global _struct_count
        uniq_name = "struct_" + str(_struct_count)
        _struct_count += 1

        field_list = []
        for i, e in enumerate(t.elements):
            # llvmlite modules get _unique string only works for symbol names
            field_uniq_name = uniq_name + "field_" + str(i)
            field_list.append((field_uniq_name, _convert_llvm_ir_to_ctype(e)))

        ret_t = type(uniq_name, (ctypes.Structure,), {"__init__": ctypes.Structure.__init__})
        ret_t._fields_ = field_list
        assert len(ret_t._fields_) == len(t.elements)
    else:
        assert False, "Don't know how to convert LLVM type: {}".format(t)

    return ret_t



from . import modulatory
from . import pathway
from . import projection

from .modulatory import *
from .pathway import *
from .projection import *

__all__ = list(modulatory.__all__)
__all__.extend(pathway.__all__)
__all__.extend(projection.__all__)



from .regressioncfa import *
from .compositionrunner import *
__all__ = list(regressioncfa.__all__)
__all__.extend(compositionrunner.__all__)

try:
    import torch

    from .autodiffcomposition import *

    __all__.extend(autodiffcomposition.__all__)
except ImportError:
    pass



import ctypes
import numpy as np
import pytest
import random

from psyneulink.core import llvm as pnlvm

SEED = 0

@pytest.mark.benchmark(group="Mersenne Twister integer PRNG")
@pytest.mark.parametrize('mode', ['Python', 'numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=pytest.mark.cuda)])
# Python uses different algorithm so skip it in this test
def test_random_int(benchmark, mode):
    res = []
    if mode == 'Python':
        state = random.Random(SEED)
        def f():
            return state.randrange(0xffffffff)
    elif mode == 'numpy':
        # Numpy promotes elements to int64
        state = np.random.RandomState([SEED])
        def f():
            return state.randint(0xffffffff, dtype=np.int64)
    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.byref_arg_types[0]()
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32')
        out = ctypes.c_longlong()
        def f():
            gen_fun(state, out)
            return out.value
    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state_size = ctypes.sizeof(init_fun.byref_arg_types[0])
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32')
        out = np.asarray([0], dtype=np.int64)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)
        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out[0]

    res = [f(), f()]
    assert np.allclose(res, [3626764237, 1654615998])
    benchmark(f)


@pytest.mark.benchmark(group="Mersenne Twister floating point PRNG")
@pytest.mark.parametrize('mode', ['Python', 'numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=pytest.mark.cuda)])
def test_random_float(benchmark, mode):
    res = []
    if mode == 'Python':
        # Python treats every seed as array
        state = random.Random(SEED)
        def f():
            return state.random()
    elif mode == 'numpy':
        # numpy promotes elements to int64
        state = np.random.RandomState([SEED])
        def f():
            return state.random_sample()
    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.byref_arg_types[0]()
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_double')
        out = ctypes.c_double()
        def f():
            gen_fun(state, out)
            return out.value
    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state_size = ctypes.sizeof(init_fun.byref_arg_types[0])
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_double')
        out = np.asfarray([0.0], dtype=np.float64)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)
        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out[0]

    res = [f(), f()]
    assert np.allclose(res, [0.8444218515250481, 0.7579544029403025])
    benchmark(f)


@pytest.mark.benchmark(group="Marsenne Twister Normal distribution")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=pytest.mark.cuda)])
# Python uses different algorithm so skip it in this test
def test_random_normal(benchmark, mode):
    if mode == 'numpy':
        # numpy promotes elements to int64
        state = np.random.RandomState([SEED])
        def f():
            return state.normal()
    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.byref_arg_types[0]()
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_normal')
        out = ctypes.c_double()
        def f():
            gen_fun(state, out)
            return out.value
    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state_size = ctypes.sizeof(init_fun.byref_arg_types[0])
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)
        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_normal')
        out = np.asfarray([0.0], dtype=np.float64)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)
        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out[0]

    res = [f(), f()]
    assert np.allclose(res, [0.4644982638709743, 0.6202001216069017])
    benchmark(f)



import ctypes
import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm


DIM_X=1000


u = np.random.rand(DIM_X)
v = np.random.rand(DIM_X)
scalar = np.random.rand()


llvm_res = np.random.rand(DIM_X)
add_res = np.add(u, v)
sub_res = np.subtract(u, v)
mul_res = np.multiply(u, v)
smul_res = np.multiply(u, scalar)


ct_u = u.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
ct_v = v.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
ct_res = llvm_res.ctypes.data_as(ctypes.POINTER(ctypes.c_double))


@pytest.mark.benchmark(group="Hadamard")
@pytest.mark.parametrize("op, y, llvm_y, builtin, result", [
                         (np.add, v, ct_v, "__pnl_builtin_vec_add", add_res),
                         (np.subtract, v, ct_v, "__pnl_builtin_vec_sub", sub_res),
                         (np.multiply, v, ct_v, "__pnl_builtin_vec_hadamard", mul_res),
                         (np.multiply, scalar, scalar, "__pnl_builtin_vec_scalar_mult", smul_res),
                         ], ids=["ADD", "SUB", "MUL", "SMUL"])
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm)])
def test_vector_op(benchmark, op, y, llvm_y, builtin, result, mode):
    if mode == 'Python':
        res = benchmark(op, u, y)
    elif mode == 'LLVM':
        llvm_fun = pnlvm.LLVMBinaryFunction.get(builtin)
        benchmark(llvm_fun, ct_u, llvm_y, DIM_X, ct_res)
        res = llvm_res
    assert np.allclose(res, result)

@pytest.mark.benchmark(group="Sum")
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm)])
def test_vector_sum(benchmark, mode):
    if mode == 'Python':
        res = benchmark(np.sum, u)
    elif mode == 'LLVM':
        llvm_fun = pnlvm.LLVMBinaryFunction.get("__pnl_builtin_vec_sum")
        benchmark(llvm_fun, ct_u, DIM_X, ct_res)
        res = llvm_res[0]
    assert np.allclose(res, sum(u))



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# **************************************** ContrastiveHebbianMechanism *************************************************

"""

Contents
--------

  * `ContrastiveHebbian_Overview`
  * `ContrastiveHebbian_Creation`
      - `ContrastiveHebbian_Size`
      - `ContrastiveHebbian_Phases`
      - `ContrastiveHebbian_OutputPorts`
      - `ContrastiveHebbian_Learning`
      - `ContrastiveHebbian_SIMPLE_HEBBIAN`
  * `ContrastiveHebbian_Structure`
      - `ContrastiveHebbian_Input`
      - `ContrastiveHebbian_Functions`
      - `ContrastiveHebbian_Output`
  * `ContrastiveHebbian_Execution`
      - `ContrastiveHebbian_Processing`
      - `ContrastiveHebbian_Learning_Execution`
  * `ContrastiveHebbian_Class_Reference`


.. _ContrastiveHebbian_Overview:

Overview
--------

A ContrastiveHebbianMechanism is a subclass of `RecurrentTransferMechanism` that is customized for use with the
Contrastive Hebbian learning rule.  See the following references for a description of the learning rule,
its relationship to the backpropagation learning rule, and its use in connectionist networks:

  `Movellan, J. R. (1991). Contrastive Hebbian learning in the continuous Hopfield model. In Connectionist Models
  (pp. 10-17) <https://www.sciencedirect.com/science/article/pii/B978148321448150007X>`_

  `Xie, X., & Seung, H. S. (2003). Equivalence of backpropagation and contrastive Hebbian learning in a layered network.
  Neural computation, 15(2), 441-454 <https://www.mitpressjournals.org/doi/abs/10.1162/089976603762552988>`_

  `O'reilly, R. C. (2001). Generalization in interactive networks: The benefits of inhibitory competition and Hebbian
  learning. Neural computation, 13(6), 1199-1241 <https://www.mitpressjournals.org/doi/abs/10.1162/08997660152002834>`_

  `Verguts, T., & Notebaert, W. (2008). Hebbian learning of cognitive control: dealing with specific and nonspecific
  adaptation. Psychological review, 115(2), 518 <http://psycnet.apa.org/record/2008-04236-010>`_

The features and operation of a ContrastiveHebbianMechanism that differ from those of a RecurrentTransferMechanism are
described below.

.. _ContrastiveHebbian_Creation:

Creating a ContrastiveHebbianMechanism
--------------------------------------

.. _ContrastiveHebbian_Size:

*Size*
~~~~~~

The **input_size** argument of the constructor must always be specified (this is comparable to specifying the
**size** or *default_variable** arguments of other types of `Mechanism`).  If it is specified on its own,
it determines the total number of processing units.  If either the **hidden_size** and/or **target_size** arguments
are specified, then those units are treated as distinct from the input units (see `ContrastiveHebbian_Execution` for
details).

.. _ContrastiveHebbian_Phases:

*Phases*
~~~~~~~~

A ContrastiveHebbianMechanism `executes in two phases <ContrastiveHebbian_Execution>`, and has
**minus_phase_termination_condition** and **plus_phase_termination_condition** arguments, and corresponding
**minus_phase_termination_threshold** and **plus_phase_termination_threshold** arguments, that determine when the
respective phases of execution are terminated.  Other parameters can also be configured that influence processing (see
`ContrastiveHebbian_Execution`).

.. _ContrastiveHebbian_OutputPorts:

*OututPorts*
~~~~~~~~~~~~

The Mechanism is automatically assigned three of its five `standard_output_ports
<ContrastiveHebbianMechanism.standard_output_ports>`: *OUTPUT_ACTIVITY*, *CURRENT_ACTIVITY*,
and *ACTIVITY_DIFFERENT* (see `below <ContrastiveHebbian_Output>`). Additional OutputPorts can be specified
in the **additional_output_ports** argument of the constructor.

.. _ContrastiveHebbian_Learning:

*Learning*
~~~~~~~~~~

A ContrastiveHebbianMechanism can be configured for learning either by specifying **enable_learning** as `True` or using
the `configure_learning <RecurrentTransferMechanism.configure_learning>` method, with the following differences from a
standard `RecurrentTransferMechanism <RecurrentTransferMechanism_Learning>`:  it is automatically assigned
`ContrastiveHebbian` as its `learning_function <ContrastiveHebbianMechanism.learning_function>`; its `learning_condition
<RecurrentTransferMechanism.learning_condition>` is automatically assigned as *CONVERGENCE*; and it is assigned a
`MappingProjection` from its *ACTIVITY_DIFFERENCE* (rather than its `primary <OutputPort_Primary>`)
`OutputPort <ContrastiveHebbian_Output>` to the *ACTIVATION_INPUT* of its `learning_mechanism
<ContrastiveHebbianMechanism.learning_mechanism>`.

.. _ContrastiveHebbian_SIMPLE_HEBBIAN:

*SIMPLE_HEBBIAN mode*
~~~~~~~~~~~~~~~~~~~~~

This replicates the function of a standard RecurrentTransferMechanism using Hebbian Learning (e.g., for validation or
ease in comparing processing outcomes).  It is configured by specifying the **mode** argument as *SIMPLE_HEBBIAN*,
which automatically makes the following assignments:

|    **separated** = `False`;
|    **clamp** = `SOFT_CLAMP`;
|    **learning_function** = `Hebbian`.
|
These assignments override any others made in the constructor, and the **hidden_size** and **target_size** arguments
are ignored.


.. _ContrastiveHebbian_Structure:

Structure
---------

.. _ContrastiveHebbian_Input:

*Input*
~~~~~~~

A ContrastiveHebbianMechanism always has two, and possibly three `InputPorts <InputPort>`: 

    * *INPUT:* receives external input to the Mechanism;
    ..
    * *RECURRENT:* receives the `value <Projection_Base.value>` of the Mechanism's `recurrent_projection
      <ContrastiveHebbianMechanism.recurrent_projection>`;
    ..
    * *TARGET:* only implemented if **target_size** is specified, **separated = `True` (default), and
      mode is not `SIMPLE_HEBBIAN <ContrastiveHebbian_SIMPLE_HEBBIAN>`;  receives the `target <Run.target>`
      specified in the `run <System.run>` method of any `Composition` to which the Mechanism belongs.

The sizes of these are determined by arguments in its constructor, which generally conform to one of two
configurations.

|    **Standard Configuration** --   if **input_size**, **hidden_size** and **target_size** are all specified,
|    the sizes of the InputPorts are as follows:
|
|        *INPUT*:  size = **input_size**
|        *RECURRENT*:  size = **input_size** + **hidden_size** + **target_size**
|        *TARGET*:  size = **input_size**
|
|    **Simple Configuration** -- if **target_size** = `None` or 0,  **separated** = `False`, and/or **mode** =
|    *SIMPLE_HEBBIAN*, the sizes of the InputPorts are as follows:
|
|        *INPUT*:  size = **input_size**
|        *RECURRENT*:  size = **input_size** + **hidden_size**
|        *TARGET*:  Not implemented
|
.. note::
   If **separated** = `False` and **target_size** is specified, a *TARGET* InputPort will be created, but
   **target_size** must equal **input_size** or an error will be generated.

The values of **input_size**, **hidden_size** and **target_size** are assigned to the Mechanism's
`input_size <ContrastiveHebbianMechanism.input_size>`, `hidden_size <ContrastiveHebbianMechanism.hidden_size>`,
and `target_size <ContrastiveHebbianMechanism.target_size>` attributes, respectively.

.. _ContrastiveHebbian_Fields:

The `input_size <ContrastiveHebbianMechanism.input_size>` and `target_size <ContrastiveHebbianMechanism.target_size>`
(if **separated** is `True`) attribute(s), together with the `separated <ContrastiveHebbianMechanism>` attribute
are used to define the fields of the *RECURRENT* InputPort's `value <InputPort.value>` and the `current_activity
<ContrastiveHebbianMechanism>` attribute used for updating values during `execution <ContrastiveHebbian_Processing>`,
as follows:

    * *input_field:*  the leftmost number of elements determined by `input_size
      <ContrastiveHebbianMechanism.input_size>`;
    ..
    * *hidden_field:*  the elements of the *RECURRENT* InputPort and `current_activity <ContrastiveHebbianMechanism>`
      that are not within the *input_field* and/or *target_field*;
    ..
    * *target_field:* the rightmost number of elements determined by `target_size
      <ContrastiveHebbianMechanism.target_size>` if `separated <ContrastiveHebbianMechanism.separated>` is `True`;
      otherwise, the same as *input_field*.

.. _ContrastiveHebbian_Functions:

*Functions*
~~~~~~~~~~~

In addition to its primary `function <Mechanism_Base.function>`, if either the
`minus_phase_termination_condition <ContrastiveHebbianMechanism.minus_phase_termination_condition>` or
`plus_phase_termination_condition <ContrastiveHebbianMechanism.plus_phase_termination_condition>`
is specified as *CONVERGENCE*, then its `phase_phase_convergence_function
<ContrastiveHebbianMechanism.phase_phase_convergence_function>` is used to determine when the corresponding
`phase of execution <ContrastiveHebbian_Execution>` is complete. Its `learning_function
<ContrastiveHebbianMechanism.learning_function>` is automatically assigned as `ContrastiveHebbian`, but this can be
replaced by any function that takes two 1d arrays ("activity ports") and compares them to determine the `matrix
<MappingProjection.matrix>`  of the Mechanism's `recurrent_projection
<ContrastiveHebbianMechanism.recurrent_projection>`.  If **mode** is specified as `SIMPLE_HEBBIAN
<ContrastiveHebbian_SIMPLE_HEBBIAN>`), the default `function <Mechanism_Base.function>` is `Hebbian`,
but can be replaced by any function that takes and returns a 1d array.

.. _ContrastiveHebbian_Output:

*Output*
~~~~~~~~

A ContrastiveHebbianMechanism is automatically assigned three `OutputPorts <OutputPort>`:

* *OUTPUT_ACTIVITY:* assigned as the `primary OutputPort <OutputPort_Primary>`, and contains the pattern
  of activity the Mechanism is trained to generate.  If **target_size** is specified, then it has the same size as
  the *TARGET* InputPort;  if **target_size** is not specified or is 0, if `separated
  <ContrastiveHebbianMechanism.separated>` is `False`, or if mode is *SIMPLE_HEBBIAN*, then the size of the
  *OUTPUT_ACTIVITY* OutputPort is the same as the *INPUT* InputPort (see `ContrastiveHebbian_Input`).
..
* *CURRENT_ACTIVITY:* assigned the value of the `current_activity <ContrastiveHebbianMechanism.current_activity>`
  attribute after each `execution <ContrastiveHebbian_Execution>` of the Mechanism, which contains the activity of all
  processing units in the Mechanism (input, hidden and/or target);  has the same size as the *RECURRENT* InputPort.
..
* *ACTIVITY_DIFFERENCE:* assigned the difference between `plus_phase_activity
  <ContrastiveHebbianMechanism.plus_phase_activity>` and `minus_phase_activity
  <ContrastiveHebbianMechanism.minus_phase_activity>` at the `completion of execution
  <ContrastiveHebbian_Execution>`. If `configured for learning <ContrastiveHebbian_Learning>`, a `MappingProjection`
  is assigned from the *ACTIVITY_DIFFERENCE* `OutputPort <ContrastiveHebbian_Output>` to the
  *ACTIVATION_INPUT* of the `learning_mechanism <ContrastiveHebbianMechanism.learning_mechanism>`.

These also appear in the ContrastiveHebbianMechanism's `standard_output_ports
<ContrastiveHebbianMechanism.standard_output_ports>`, along with two others -- *MINUS_PHASE_ACTIVITY* and
*PLUS_PHASE_ACTIVITY* -- that it can be assigned, in addition to the `standard_output_ports
<RecurrentTransferMechanism.standard_output_ports>` of a `RecurrentTransferMechanism`.


.. _ContrastiveHebbian_Execution:

Execution
---------

.. _ContrastiveHebbian_Processing:

*Processing*
~~~~~~~~~~~~

A ContrastiveHebbianMechanism always executes in two sequential phases, that together constitute a *trial of execution:*

.. _ContrastiveHebbian_Minus_Phase:

* *minus phase:* in each execution, the *RECURRENT* InputPort's `value <InputPort.value>` (received from the
  `recurrent_projection <ContrastiveHebbianMechanism.recurrent_projection>`) is combined with the *INPUT*
  InputPort's `value <InputPort.value>`. The result of `function <Mechanism_Base.function>` is assigned to
  `current_activity <ContrastiveHebbianMechanism.current_activity>`.  The Mechanism is executed repeatedly
  until its `minus_phase_termination_condition <ContrastiveHebbianMechanism.minus_phase_termination_condition>`
  is met. At that point, the *minus phase* is completed, the `value <Mechanism_Base.value>` of the
  ContrastiveHebbianMechanism is assigned to its `minus_phase_activity
  <ContrastiveHebbianMechanism.minus_phase_activity>` attribute, and the *plus phase* is begun.
..

.. _ContrastiveHebbian_Plus_Phase:

* *plus phase:*  if `continuous <ContrastiveHebbianMechanism.continuous>` is `False`, then `current_activity
  <ContrastiveHebbianMechanism.current_activity>` is reset to
  `initial_value <Mechanism_Base.initial_value>`, and the Mechanism's
  previous `value <Mechanism_Base.value>` is reset to ``None``;
  otherwise, these retain their value from the last execution in the
  *minus phase*.  In either case, the *RECURRENT* InputPort's `value <InputPort.value>` is combined with the *INPUT*
  InputPort's `value <InputPort.value>` (as during the `minus_phase
  <ContrastiveHebbianMechanism.minus_phase_activity>`) as well as that of *TARGET* InputPort (if that is `specified
  <ContrastiveHebbian_Input>`). If `separated <ContrastiveHebbianMechanism.separated>` is `True` (which it is by
  default), then the `value <InputPort.value>` of the *INPUT* `InputPort <ContrastiveHebbian_Input>` is combined with
  the `input_field <ContrastiveHebbian_Fields>` of the *RECURRENT* InputPort's `value <InputPort.value>`,
  and *TARGET* is combined with the `target_field <ContrastiveHebbian_Fields>`; otherwise, both are combined with the
  `input_field <ContrastiveHebbian_Fields>`.  If `hidden_size <ContrastiveHebbianMechanism.hidden_size>` is
  specified, then the `hidden_field <ContrastiveHebbian_Fields>` of the *RECURRENT* InputPort's  `value
  <InputPort.value>` is determined only by the `value <AutoAssociativeProjection.value>` of the Mechanism's
  `recurrent_projection <ContrastiveHebbianMechanism.recurrent_projection>`.  Execution then proceeds as during the
  *minus phase*, completing when its `plus_phase_termination_condition
  <ContrastiveHebbianMechanism.plus_phase_termination_condition>` is met.  At that point, the *plus phase* is
  completed, and the `value <Mechanism_Base.value>` of the Mechanism is assigned to
  `plus_phase_activity <ContrastiveHebbianMechanism.minus_phase_activity>`.

The `value <InputPort.value>` of the *INPUT*, and possibly *TARGET*, InptutState(s) are combined with that of its
*RECURRENT* InputPort using the `combination_function <ContrastiveHebbianMechanism.combination_function>`. The
manner in which these are combined is determined by the `clamp <ContrastiveHebbianMechanism.clamp>` attribute: if
it is *HARD_CLAMP* they are used to replace the corresponding fields of *RECURRENT*;  if it is *SOFT_CLAMP*, *INPUT*
(and possibly *TARGET*) are added to *RECURRENT*; .  The result is passed to the Mechanism's `integrator_function
<TransferMechanism.integrator_function>` (if `integrator_mode <TransferMechanism.integrator_mode>`
is `True`) and then its `function <Mechanism_Base.function>`.

If the termination condition for either phase is specified as *CONVERGENCE*, it uses the Mechanism's
`phase_convergence_function <ContrastiveHebbianMechanism.phase_convergence_function>`, together with the
termination_threshold specified for that phase, to determine when execution of that phase terminates.  If `max_passes
<ContrastiveHebbianMechanism.max_passes>` is specified, and the number of executions in either phase
reaches that value, an error is generated.  Otherwise, once a trial of execution is complete (i.e, after completion
of the *minus phase*), the following computations and assignments are made:

* if a *TARGET* InputPort `has been specified <ContrastiveHebbian_Input>`, then the `target field
  <ContrastiveHebbian_Fields>` of `current_activity <ContrastiveHebbianMechanism.current_activity>` is assigned as
  `value <OutputPort.value>` of *OUTPUT_ACTIVITY* `OutputPort <ContrastiveHebbian_Output>`;  otherwise,
  it is assigned the value of the `input_field <ContrastiveHebbian_Fields>` of `current_activity
  <ContrastiveHebbianMechanism.current_activity>`.
..
* `plus_phase_activity <ContrastiveHebbianMechanism.plus_phase_activity>` is assigned as the `value
  <OutputPort.value>` of the *CURRENT_ACTIVITY_ATTR* `OutputPort <ContrastiveHebbian_Output>`;
..
* the difference between `plus_phase_activity <ContrastiveHebbianMechanism.plus_phase_activity>` and
  `minus_phase_activity <ContrastiveHebbianMechanism.minus_phase_activity>` is assigned as the `value
  <OutputPort.value>` of the *ACTIVITY_DIFFERENCE* `OutputPort <ContrastiveHebbian_Output>`.

.. _ContrastiveHebbian_Learning_Execution:

*Learning*
~~~~~~~~~~

If a ContrastiveHebbianMechanism is `configured for learning <ContrastiveHebbian_Learning>`, at the end of each
`trial of execution <ContrastiveHebbian_Processing>` the `value <OutputPort.value>` of its *ACTIVITY_DIFFERENCE*
`OutputPort <ContrastiveHebbian_Output>` is passed to its `learning_mechanism
<ContrastiveHebbianMechanism.learning_mechanism>`.  If the Mechanism is part of a `System`, then the
`learning_mechanism <ContrastiveHebbianMechanism.learning_mechanism>` is executed during the `execution phase
<System_Execution>` of the System's execution.  Note that this is distinct from the behavior of supervised learning
algorithms (such as `Reinforcement` and `BackPropagation`), that are executed during the `learning phase
<System_Execution>` of a System's execution

.. _ContrastiveHebbian_Class_Reference:

Class Reference
---------------

"""

from collections.abc import Iterable

import copy
import numpy as np
import typecheck as tc
from psyneulink.core.components.functions.function import get_matrix, is_function_type
from psyneulink.core.components.functions.learningfunctions import ContrastiveHebbian, Hebbian
from psyneulink.core.components.functions.objectivefunctions import Distance
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import \
    CONTRASTIVE_HEBBIAN_MECHANISM, COUNT, FUNCTION, HARD_CLAMP, HOLLOW_MATRIX, MAX_ABS_DIFF, NAME, \
    SIZE, SOFT_CLAMP, TARGET, VARIABLE
from psyneulink.core.globals.parameters import Parameter, SharedParameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.utilities import is_numeric_or_none, parameter_spec
from psyneulink.library.components.mechanisms.processing.transfer.recurrenttransfermechanism import \
    CONVERGENCE, RECURRENT, RECURRENT_INDEX, RecurrentTransferMechanism
from psyneulink.library.components.projections.pathway.autoassociativeprojection import AutoAssociativeProjection

__all__ = [
    'ACTIVITY_DIFFERENCE', 'CURRENT_ACTIVITY', 'SIMPLE_HEBBIAN', 'INPUT',
    'MINUS_PHASE_ACTIVITY', 'PLUS_PHASE_ACTIVITY', 'OUTPUT_ACTIVITY',
    'ContrastiveHebbianError', 'ContrastiveHebbianMechanism',
]

INPUT = 'INPUT'

INPUT_SIZE = 'input_size'
HIDDEN_SIZE = 'hidden_size'
TARGET_SIZE = 'target_size'
SEPARATED = 'separated'

SIMPLE_HEBBIAN = 'SIMPLE_HEBBIAN'

OUTPUT_ACTIVITY = 'OUTPUT_ACTIVITY'
CURRENT_ACTIVITY = 'CURRENT_ACTIVITY'
ACTIVITY_DIFFERENCE = 'ACTIVITY_DIFFERENCE'
MINUS_PHASE_ACTIVITY = 'MINUS_PHASE_ACTIVITY'
PLUS_PHASE_ACTIVITY = 'PLUS_PHASE_ACTIVITY'

OUTPUT_ACTIVITY_ATTR = 'output_activity'
CURRENT_ACTIVITY_ATTR = 'current_activity'
MINUS_PHASE_ACTIVITY_ATTR = 'minus_phase_activity'
PLUS_PHASE_ACTIVITY_ATTR = 'plus_phase_activity'

INPUT_INDEX = 0
RECURRENT_INDEX = 1
TARGET_INDEX = 2

MINUS_PHASE = False
PLUS_PHASE  = True


class ContrastiveHebbianError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


def _CHM_output_activity_getter(owning_component=None, context=None):
    current_activity = owning_component.parameters.current_activity._get(context)
    if owning_component.target_size:
        return current_activity[owning_component.target_start:owning_component.target_end]
    else:
        return current_activity[:owning_component.input_size]

def _CHM_input_activity_getter(owning_component=None, context=None):
    current_activity = owning_component.parameters.current_activity._get(context)
    return current_activity[:owning_component.input_size]


def _CHM_hidden_activity_getter(owning_component=None, context=None):
    if owning_component.hidden_size:
        current_activity = owning_component.parameters.current_activity._get(context)
        return current_activity[owning_component.input_size:owning_component.target_start]


def _CHM_target_activity_getter(owning_component=None, context=None):
    if owning_component.target_size:
        current_activity = owning_component.parameters.current_activity._get(context)
        return current_activity[owning_component.target_start:owning_component.target_end]


class ContrastiveHebbianMechanism(RecurrentTransferMechanism):
    """
    ContrastiveHebbianMechanism(                                          \
                input_size=2,                                             \
                hidden_size=None,                                         \
                target_size=None,                                         \
                separated:bool=True,                                      \
                mode=None,                                                \
                continuous=True,                                          \
                clamp=HARD_CLAMP,                                         \
                minus_phase_termination_condition = CONVERGENCE,          \
                minus_phase_termination_threshold=.01,                    \
                plus_phase_termination_condition = CONVERGENCE,           \
                plus_phase_termination_threshold=.01,                     \
                phase_convergence_function=Distance(metric=MAX_ABS_DIFF), \
                max_passes=None,                                          \
                additional_input_ports=None,                              \
                additional_output_ports=None)

    Subclass of `RecurrentTransferMechanism` that implements a single-layer auto-recurrent network using two-phases
    of execution and the `Contrastive Hebbian Learning algorithm
    <https://www.sciencedirect.com/science/article/pii/B978148321448150007X>`_.  See `RecurrentTransferMechanism
    <RecurrentTransferMechanism_Class_Reference>` for additional arguments and attributes.

    Arguments
    ---------

    input_size : int : default 0
        specifies the size of the *INPUT* `InputPort <ContrastiveHebbian_Input>` and number of units in the
        `input_field <ContrastiveHebbian_Fields>` of `current_activity <ContrastiveHebbianMechanism.current_activity>`.

    hidden_size : int : default None
        specifies the number of units in the `hidden_field <ContrastiveHebbian_Fields>` of `current_activity
        <ContrastiveHebbianMechanism.current_activity>`.

    target_size : int : default None
        specifies the size of the *TARGET* `InputPort <ContrastiveHebbian_Input>`, number of units in the
        `input_field <ContrastiveHebbian_Fields>` of `current_activity <ContrastiveHebbianMechanism.current_activity>`,
        and of the `OUTPUT_ACTIVITY` `OutputPort <ContrastiveHebbian_Output>`.

    separated : bool : default True
        specifies whether `target_field <ContrastiveHebbian_Fields>` should be different from the `input_field
        <ContrastiveHebbian_Fields>`.

    mode : SIMPLE_HEBBIAN or None : default None
        specifies configuratoin that emulates standard `RecurrentTransferMechanism` using `Hebbian` learning
        (see `SIMPLE_HEBBIAN mode <ContrastiveHebbian_SIMPLE_HEBBIAN>` for details).

    combination_function : function : default None
        specifies function used to combine the *INPUT*, *RECURRENT* and *TARGET* (if specified) `InputPorts
        <ContrastiveHebbian_Input>`; must accept a 2d array with two or three items of the same length, and generate a
        result that is the same size as `recurrent_size <ContrastiveHebbianMechanism.recurrent_size>`;  if `None`,
        the ContrastiveHebbianMechanism's combination_function method is used.

    clamp : HARD_CLAMP or SOFT_CLAMP : default HARD_CLAMP
        specifies the manner in which the `value <InputPort.value>` of the  *INPUT* and *TARGET* (if specified)
        `InputPorts <ContrastiveHebbian_Input>` are combined with the *RECURRENT* `InputPort
        <ContrastiveHebbian_Input>` in each execution (see `clamp <ContrastiveHebbianMechanism.clamp>` for additional
        details.

    continuous : bool : default True
        specifies whether or not to reset `current_activity <ContrastiveHebbianMechanism.current_activity>`
        at the beginning of the `minus phase <ContrastiveHebbian_Minus_Phase>` of a trial.

    minus_phase_termination_condition : COUNT or CONVERGENCE : default CONVERGENCE
        specifies the type of condition used to terminate the `minus_phase <ContrastiveHebbian_Minus_Phase>` of
        execution (see `minus_phase_termination_condition
        <ContrastiveHebbianMechanism.minus_phase_termination_condition>` for additional details).

    minus_phase_termination_threshold : float or int : default 0.01
        specifies the value used to determine when the `minus_phase <ContrastiveHebbian_Minus_Phase>` terminates;
        should be a float if `minus_phase_termination_condition
        <ContrastiveHebbianMechanism.minus_phase_termination_condition>` is *CONVERGENCE*, and int if it is *COUNT*
        (see `minus_phase_termination_threshold <ContrastiveHebbianMechanism.minus_phase_termination_threshold>`
        for additional details).

    plus_phase_termination_condition : COUNT or CONVERGENCE : default CONVERGENCE
        specifies the type of condition used to terminate the `plus_phase <ContrastiveHebbian_Plus_Phase>` of
        execution (see `plus_phase_termination_condition <ContrastiveHebbianMechanism.plus_phase_termination_condition>`
        for additional details).

    plus_phase_termination_threshold : float or int : default 0.01
        specifies the value used to determine when the `minus_phase <ContrastiveHebbian_Plus_Phase>` terminates;
        should be a float if `plus_phase_termination_condition
        <ContrastiveHebbianMechanism.plus_phase_termination_condition>` is *CONVERGENCE*, and int if it is *COUNT*
        (see `plus_phase_termination_threshold <ContrastiveHebbianMechanism.plus_phase_termination_threshold>`
        for additional details).

    phase_convergence_function : function : default Distance(metric=MAX_ABS_DIFF)
        specifies the function that determines when a `phase of execution <ContrastiveHebbian_Execution>` is complete
        if the termination condition for that phase is specified as *CONVERGENCE*, by comparing `current_activity
        <ContrastiveHebbianMechanism.current_activity>` with the previous `value
        <Mechanism_Base._value>` of the Mechanism;  can be any function that takes two 1d arrays
        of the same length as `variable <Mechanism_Base.variable>` and returns a scalar value. The default
        is the `Distance` Function, using the `MAX_ABS_DIFF` metric  which computes the elementwise difference between
        two arrays and returns the difference with the maximum absolute value.

    max_passes : int : default 1000
        specifies maximum number of executions (`passes <TimeScale.PASS>`) that can occur in an `execution phase
        <ContrastiveHebbian_Execution>` before reaching the termination_threshold for that phase,
        after which an error occurs; if `None` is specified, execution may continue indefinitely or until an
        interpreter exception is generated.

    Attributes
    ----------

    input_size : int
        size of the *INPUT* `InputPort <ContrastiveHebbian_Input>` and `input_activity
        <ContrastiveHebbianMechanism.input_activity>`, and the number of units in the `input_field of
        <ContrastiveHebbian_Fields>` of `current_activity <ContrastiveHebbianMechanism.current_activity>`.

    hidden_size : int
        size of `hidden_activity <ContrastiveHebbianMechanism.input_activity>`, and number of units in
        `current_activity <ContrastiveHebbianMechanism.current_activity>` and the `hidden_field
        <ContrastiveHebbian_Fields>` of the *RECURRENT* `InputPort <ContrastiveHebbian_Input>`.

    target_size : int
        size of the *TARGET* `InputPort <ContrastiveHebbian_Input>` `if specified <ContrastiveHebbian_Creation>` and,
        if so, the number of units in `target_activity <ContrastiveHebbianMechanism.target_activity>`, the
        `target_field <ContrastiveHebbian_Fields>` of `current_activity <ContrastiveHebbianMechanism.current_activity>`,
        and the *OUTPUT_ACTIVITY* `OutputPort <ContrastiveHebbian_Output>`.

    target_start : int
        index of first unit of `target_field <ContrastiveHebbian_Fields>`.

    target_end : int
        index of first unit *after* `target_field <ContrastiveHebbian_Fields>`.

    separated : bool : default True
        determines whether `target_field <ContrastiveHebbian_Fields>` is different from `input_field
        <ContrastiveHebbian_Fields>` (`True`) or the same (`False`).

    recurrent_size : int
        size of *RECURRENT* `InputPort <ContrastiveHebbian_Input>`, `current_activity
        <ContrastiveHebbianMechanism.current_activity>`, and *CURRENT_ACTIVITY* `OutputPort
        <ContrastiveHebbian_Output>`.

    mode : SIMPLE_HEBBIAN or None
        indicates whether *SIMPLE_HEBBIAN* was used for configuration (see
        `SIMPLE_HEBBIAN mode <ContrastiveHebbian_SIMPLE_HEBBIAN>` for details).

    recurrent_projection : AutoAssociativeProjection
        an `AutoAssociativeProjection` that projects from the *CURRENT_ACTIVITY* `OutputPort
        <ContrastiveHebbian_Output>` to the *RECURRENT* `InputPort <ContrastiveHebbian_Input>`.

    combination_function : method or function
        used to combine `value <InputPort.value>` of the *INPUT* and *TARGET* (if specified)  `InputPorts
        <ContrastiveHebbian_Input>` with that of the *RECURRENT* `InputPort <ContrastiveHebbian_Input>` to determine
        the `variable <CurrentHebbianMechanism.variable>` passed to the Mechanism's `integrator_function
        <TransferMechanism.integrator_function>` and/or its `function <Mechanism_Base.function>`
        (see `ContrastiveHebbian_Execution` for details).

    clamp : HARD_CLAMP or SOFT_CLAMP
        determines whether the `value <InputPort.value>` of the *INPUT* and *TARGET* (if specified) `InputPorts
        <ContrastiveHebbian_Input>` replace (*HARD_CLAMP*) or are added to (*SOFT_CLAMP*) the `value <InputPort.value>`
        of the *RECURRENT* InputPort by `combination_function <ContrastiveHebbianMechanism.combination_function>`.

    continuous : bool : default True
        determines whether or not `current_activity <ContrastiveHebbianMechanism.current_activity>` is reset
        at the beginning of the `minus phase <ContrastiveHebbian_Minus_Phase>` of execution. If `False`, it is set to `initial_value
        <Mechanism_Base.initial_value>`.

    current_activity : 1d array of floats
        the value of the actvity of the ContrastiveHebbianMechanism following its last `execution
        <ContrastiveHebbian_Execution>`.

    input_activity : 1d array of floats
        value of units in the `input_field <ContrastiveHebbian_Fields>` of `current_activity
        <ContrastiveHebbianMechanism.current_activity>`.

    hidden_activity : 1d array of floats or None
        value of units in the `hidden_field <ContrastiveHebbian_Fields>` of `current_activity
        <ContrastiveHebbianMechanism.current_activity>`;  `None` if hidden_size = 0 or mode = `SIMPLE_HEBBIAN
        <ContrastiveHebbian_SIMPLE_HEBBIAN>`.

    target_activity : 1d array of floats or None
        value of units in the `target_field <ContrastiveHebbian_Fields>` of `current_activity
        <ContrastiveHebbianMechanism.current_activity>` if *TARGET* `InputPort is specified
        <ContrastiveHebbian_Input>`. Same as `input_activity <ContrastiveHebbianMechanism.input_activity>`
        if `separated <ContrastiveHebbianMechanism.separated>` is `False`;  `None` if target_size = 0 or mode =
        `SIMPLE_HEBBIAN <ContrastiveHebbian_SIMPLE_HEBBIAN>`.

    recurrent_activity : 1d array of floats
        same as `current_activity <ContrastiveHebbianMechanism.current_activity>`

    minus_phase_activity : 1d array of floats
        the value of the `current_activity <ContrastiveHebbianMechanism.current_activity>` at the end of the
        `minus phase of execution <ContrastiveHebbian_Minus_Phase>`.

    plus_phase_activity : 1d array of floats
        the value of the `current_activity <ContrastiveHebbianMechanism.current_activity>` at the end of the
        `plus phase of execution <ContrastiveHebbian_Plus_Phase>`.

    delta : scalar
        value returned by `phase_convergence_function <RecurrentTransferMechanism.phase_convergence_function>`;
        used to determined when `is_converged <RecurrentTransferMechanism.is_converged>` is `True`.

    is_converged : bool
        indicates when a `phase of execution <ContrastiveHebbian_Execution>` is complete, if the termination condition
        for that phase is specified as *CONVERGENCE*.  `True` when `delta <ContrastiveHebbianMechanism.delta>` is less
        than or equal to the termination_threshold speified for the corresponding phase.

    phase_convergence_function : function
        determines when a `phase of execution <ContrastiveHebbian_Execution>` is complete if the termination
        condition for that phase is specified as *CONVERGENCE*.  Compares the value of `current_activity
        <ContrastiveHebbianMechanism.current_activity>` with the
        previous `value <Mechanism_Base.value>`; result is
        assigned as the value of `delta
        <ContrastiveHebbianMechanism.delta>.

    minus_phase_termination_condition : CONVERGENCE or COUNT: default CONVERGENCE
        determines the type of condition used to terminate the `minus_phase <ContrastiveHebbian_Minus_Phase>` of
        execution.  If it is *CONVERGENCE*, execution continues until the value returned by `phase_convergence_function
        <ContrastiveHebbianMechanism.phase_convergence_function>` is less than or equal to
        `minus_phase_termination_threshold <ContrastiveHebbianMechanism.minus_phase_termination_threshold>`;
        if it is *COUNT*, the Mechanism is executed the number of times specified in
        `minus_phase_termination_threshold <ContrastiveHebbianMechanism.minus_phase_termination_threshold>`.

    minus_phase_termination_threshold : float or int
        the value used for the specified `minus_phase_termination_condition
        <ContrastiveHebbianMechanism.minus_phase_termination_condition>` to determine when the
        `minus phase of execution <ContrastiveHebbian_Minus_Phase>` terminates.

    plus_phase_termination_condition : CONVERGENCE or COUNT : default CONVERGENCE
        determines the type of condition used to terminate the `plus_phase <ContrastiveHebbian_Plus_Phase>` of
        execution.  If it is *CONVERGENCE*, execution continues until the value returned by `phase_convergence_function
        <ContrastiveHebbianMechanism.phase_convergence_function>` is less than or equal to
        `plus_phase_termination_threshold <ContrastiveHebbianMechanism.plus_phase_termination_threshold>`;
        if it is *COUNT*, the Mechanism is executed the number of times specified in
        `plus_phase_termination_threshold <ContrastiveHebbianMechanism.plus_phase_termination_threshold>`.

    plus_phase_termination_threshold : float or int
        the value used for the specified `plus_phase_termination_condition
        <ContrastiveHebbianMechanism.plus_phase_termination_condition>` to determine when the `plus phase of execution
        <ContrastiveHebbian_Plus_Phase>` terminates.

    max_passes : int or None
        determines the maximum number of executions (`passes <TimeScale.PASS>`) that can occur in an `execution phase
        <ContrastiveHebbian_Execution>` before reaching the corresponding termination_threshold for that phase,
        after which an error occurs; if `None` is specified, execution may continue indefinitely or until an
        interpreter exception is generated.

    execution_phase : bool
        indicates current `phase of execution <ContrastiveHebbian_Execution>`.
        `False` = `minus phase <ContrastiveHebbian_Minus_Phase>`;
        `True` = `plus phase <ContrastiveHebbian_Plus_Phase>`.

    output_ports : ContentAddressableList[OutputPort]
        contains the following `OutputPorts <OutputPort>` by default:

        * *OUTPUT_ACTIVITY*, the `primary OutputPort  <OutputPort_Primary>` of the Mechanism, the `value
          <OutputPort.value>` of which is `target_activity <ContrastiveHebbianMechanism.target_activity>` if a
          *TARGET* `InputPort <ContrastiveHebbian_Input>` is implemented;  otherwise,  `input_activity
          <ContrastiveHebbianMechanism.input_activity>` is assigned as its `value <OutputPort.value>`.

        * *CURRENT_ACTIVITY* -- the `value <OutputPort.value>` of which is a 1d array containing the activity
          of the ContrastiveHebbianMechanism after each execution;  at the end of an `execution sequence
          <ContrastiveHebbian_Execution>`, it is assigned the value of `plus_phase_activity
          <ContrastiveHebbianMechanism.plus_phase_activity>`.

        * *ACTIVITY_DIFFERENCE*, the `value <OutputPort.value>` of which is a 1d array with the elementwise
          differences in activity between the plus and minus phases at the end of an `execution sequence
          <ContrastiveHebbian_Execution>`.

    standard_output_ports : list[str]
        list of `Standard OutputPorts <OutputPort_Standard>` that includes the following in addition to the
        `standard_output_ports <RecurrentTransferMechanism.standard_output_ports>` of a `RecurrentTransferMechanism`:

        .. _OUTPUT_ACTIVITY:

        *OUTPUT_ACTIVITY* : 1d np.array
            array with activity of the `target_field <ContrastiveHebbian_Fields>` of `current_activity
            <ContrastiveHebbianMechanism.current_activity>` if a *TARGET* `InputPort is specified
            <ContrastiveHebbian_Input>`;  otherwise, has activity of the `input_field <ContrastiveHebbian_Fields>` of
            `current_activity <ContrastiveHebbianMechanism.current_activity>`.

        .. _CURRENT_ACTIVITY:

        *CURRENT_ACTIVITY* : 1d np.array
            array with `current_activity <ContrastiveHebbianMechanism.current_activity>`.

        .. _ACTIVITY_DIFFERENCE:

        *ACTIVITY_DIFFERENCE* : 1d np.array
            array of element-wise differences between `plus_phase_activity
            <ContrastiveHebbianMechanism.plus_phase_activity>` and `minus_phase_activity
            <ContrastiveHebbianMechanism.minus_phase_activity>`.

        .. _MINUS_PHASE_ACTIVITY:

        *MINUS_PHASE_ACTIVITY* : 1d np.array
            array `minus_phase_activity <ContrastiveHebbianMechanism.minus_phase_activity>`
            (i.e., activity at the end of the `minus phase of execution <ContrastiveHebbian_Minus_Phase>`.

        .. _PLUS_PHASE_ACTIVITY:

        *PLUS_PHASE_ACTIVITY* : 1d np.array
            array `plus_phase_activity <ContrastiveHebbianMechanism.plus_phase_activity>`
            (i.e., activity at the end of the `plus phase of execution <ContrastiveHebbian_Plus_Phase>`.


    Returns
    -------
    instance of ContrastiveHebbianMechanism : ContrastiveHebbianMechanism

    """
    componentType = CONTRASTIVE_HEBBIAN_MECHANISM

    class Parameters(RecurrentTransferMechanism.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <ContrastiveHebbianMechanism.variable>`

                    :default value: numpy.array([[0, 0]])
                    :type: ``numpy.ndarray``

                clamp
                    see `clamp <ContrastiveHebbianMechanism.clamp>`

                    :default value: `HARD_CLAMP`
                    :type: ``str``

                combination_function
                    see `combination_function <ContrastiveHebbianMechanism.combination_function>`

                    :default value: None
                    :type:

                continuous
                    see `continuous <ContrastiveHebbianMechanism.continuous>`

                    :default value: True
                    :type: ``bool``

                current_activity
                    see `current_activity <ContrastiveHebbianMechanism.current_activity>`

                    :default value: None
                    :type:

                current_termination_condition
                    see `current_termination_condition <ContrastiveHebbianMechanism.current_termination_condition>`

                    :default value: None
                    :type:

                current_termination_threshold
                    see `current_termination_threshold <ContrastiveHebbianMechanism.current_termination_threshold>`

                    :default value: None
                    :type:

                execution_phase
                    see `execution_phase <ContrastiveHebbianMechanism.execution_phase>`

                    :default value: None
                    :type:
                    :read only: True

                hidden_activity
                    see `hidden_activity <ContrastiveHebbianMechanism.hidden_activity>`

                    :default value: None
                    :type:
                    :read only: True

                hidden_size
                    see `hidden_size <ContrastiveHebbianMechanism.hidden_size>`

                    :default value: 0
                    :type: ``int``

                input_activity
                    see `input_activity <ContrastiveHebbianMechanism.input_activity>`

                    :default value: None
                    :type:
                    :read only: True

                input_size
                    see `input_size <ContrastiveHebbianMechanism.input_size>`

                    :default value: None
                    :type:

                learning_function
                    see `learning_function <ContrastiveHebbianMechanism.learning_function>`

                    :default value: `ContrastiveHebbian`
                    :type: `Function`

                max_passes
                    see `max_passes <ContrastiveHebbianMechanism.max_passes>`

                    :default value: 1000
                    :type: ``int``

                minus_phase_activity
                    see `minus_phase_activity <ContrastiveHebbianMechanism.minus_phase_activity>`

                    :default value: None
                    :type:

                minus_phase_termination_condition
                    see `minus_phase_termination_condition <ContrastiveHebbianMechanism.minus_phase_termination_condition>`

                    :default value: `CONVERGENCE`
                    :type: ``str``

                minus_phase_termination_threshold
                    see `minus_phase_termination_threshold <ContrastiveHebbianMechanism.minus_phase_termination_threshold>`

                    :default value: 0.01
                    :type: ``float``

                mode
                    see `mode <ContrastiveHebbianMechanism.mode>`

                    :default value: None
                    :type:

                output_activity
                    see `output_activity <ContrastiveHebbianMechanism.output_activity>`

                    :default value: None
                    :type:
                    :read only: True

                phase_convergence_function
                    see `phase_convergence_function <ContrastiveHebbianMechanism.phase_convergence_function>`

                    :default value: `Distance`
                    :type: `Function`

                phase_convergence_threshold
                    internal parameter, used by is_converged;  assigned the value of the termination_threshold for
                    the current `phase of execution <ContrastiveHebbian_Execution>`.

                    :default value: 0.01
                    :type: ``float``

                phase_execution_count
                    see `phase_execution_count <ContrastiveHebbianMechanism.phase_execution_count>`

                    :default value: 0
                    :type: ``int``

                phase_terminated
                    see `phase_terminated <ContrastiveHebbianMechanism.phase_terminated>`

                    :default value: False
                    :type: ``bool``

                plus_phase_activity
                    see `plus_phase_activity <ContrastiveHebbianMechanism.plus_phase_activity>`

                    :default value: None
                    :type:

                plus_phase_termination_condition
                    see `plus_phase_termination_condition <ContrastiveHebbianMechanism.plus_phase_termination_condition>`

                    :default value: `CONVERGENCE`
                    :type: ``str``

                plus_phase_termination_threshold
                    see `plus_phase_termination_threshold <ContrastiveHebbianMechanism.plus_phase_termination_threshold>`

                    :default value: 0.01
                    :type: ``float``

                separated
                    see `separated <ContrastiveHebbianMechanism.separated>`

                    :default value: True
                    :type: ``bool``

                target_activity
                    see `target_activity <ContrastiveHebbianMechanism.target_activity>`

                    :default value: None
                    :type:
                    :read only: True

                target_size
                    see `target_size <ContrastiveHebbianMechanism.target_size>`

                    :default value: 0
                    :type: ``int``
        """
        variable = np.array([[0, 0]])
        current_activity = Parameter(None, aliases=['recurrent_activity'])
        plus_phase_activity = None
        minus_phase_activity = None
        current_termination_condition = None
        current_termination_threshold = None
        phase_execution_count = 0
        phase_terminated = False

        input_size = Parameter(None, stateful=False, loggable=False)
        hidden_size = Parameter(0, stateful=False, loggable=False)
        target_size = Parameter(0, stateful=False, loggable=False)
        separated = Parameter(True, stateful=False, loggable=False)
        mode = Parameter(None, stateful=False, loggable=False)
        continuous = Parameter(True, stateful=False, loggable=False)
        clamp = Parameter(HARD_CLAMP, stateful=False, loggable=False)
        combination_function = Parameter(None, stateful=False, loggable=False)
        phase_convergence_function = Parameter(Distance(metric=MAX_ABS_DIFF), stateful=False, pnl_internal=True, loggable=False)
        phase_convergence_threshold = Parameter(0.01, modulable=True, pnl_internal=True, loggable=False)

        minus_phase_termination_condition = Parameter(CONVERGENCE, stateful=False, loggable=False)
        plus_phase_termination_condition = Parameter(CONVERGENCE, stateful=False, loggable=False)
        learning_function = SharedParameter(
            ContrastiveHebbian,
            attribute_name='learning_mechanism',
            shared_parameter_name='function',
        )
        max_passes = Parameter(1000, stateful=False)

        output_activity = Parameter(None, read_only=True, getter=_CHM_output_activity_getter)
        input_activity = Parameter(None, read_only=True, getter=_CHM_input_activity_getter)
        hidden_activity = Parameter(None, read_only=True, getter=_CHM_hidden_activity_getter)
        target_activity = Parameter(None, read_only=True, getter=_CHM_target_activity_getter)

        execution_phase = Parameter(None, read_only=True)
        # is_finished_ = Parameter(False, read_only=True)

        minus_phase_termination_threshold = Parameter(0.01, modulable=True)
        plus_phase_termination_threshold = Parameter(0.01, modulable=True)


    standard_output_ports = RecurrentTransferMechanism.standard_output_ports.copy()
    standard_output_ports.extend([{NAME:OUTPUT_ACTIVITY,
                                    VARIABLE:OUTPUT_ACTIVITY_ATTR},
                                   {NAME:CURRENT_ACTIVITY,
                                    VARIABLE:CURRENT_ACTIVITY_ATTR},
                                   {NAME:ACTIVITY_DIFFERENCE,
                                    VARIABLE:[PLUS_PHASE_ACTIVITY_ATTR, MINUS_PHASE_ACTIVITY_ATTR],
                                    FUNCTION: lambda v: v[0] - v[1]},
                                   {NAME:MINUS_PHASE_ACTIVITY,
                                    VARIABLE:MINUS_PHASE_ACTIVITY_ATTR},
                                   {NAME:PLUS_PHASE_ACTIVITY,
                                    VARIABLE:PLUS_PHASE_ACTIVITY_ATTR},
                                   ])
    standard_output_port_names = RecurrentTransferMechanism.standard_output_port_names.copy()
    standard_output_port_names = [i['name'] for i in standard_output_ports]

    @tc.typecheck
    def __init__(self,
                 input_size:int,
                 hidden_size:tc.optional(int)=None,
                 target_size:tc.optional(int)=None,
                 separated: tc.optional(bool) = None,
                 mode:tc.optional(tc.enum(SIMPLE_HEBBIAN))=None,
                 continuous: tc.optional(bool) = None,
                 clamp:tc.optional(tc.enum(SOFT_CLAMP, HARD_CLAMP))=None,
                 combination_function:tc.optional(is_function_type)=None,
                 function=None,
                 matrix=None,
                 auto=None,
                 hetero=None,
                 integrator_function=None,
                 initial_value=None,
                 noise=None,
                 integration_rate: is_numeric_or_none=None,
                 integrator_mode: tc.optional(bool) = None,
                 clip=None,
                 minus_phase_termination_condition:tc.optional(tc.enum(CONVERGENCE, COUNT))=None,
                 minus_phase_termination_threshold: tc.optional(float) = None,
                 plus_phase_termination_condition:tc.optional(tc.enum(CONVERGENCE, COUNT))=None,
                 plus_phase_termination_threshold: tc.optional(float) = None,
                 phase_convergence_function: tc.optional(tc.any(is_function_type)) = None,
                 max_passes:tc.optional(int)=None,
                 enable_learning: tc.optional(bool) = None,
                 learning_rate:tc.optional(tc.any(parameter_spec, bool))=None,
                 learning_function: tc.optional(tc.any(is_function_type)) = None,
                 additional_input_ports:tc.optional(tc.optional(tc.any(list, dict))) = None,
                 additional_output_ports:tc.optional(tc.any(str, Iterable))=None,
                 params=None,
                 name=None,
                 prefs: is_pref_set=None,
                 **kwargs):
        """Instantiate ContrastiveHebbianMechanism
        """

        if mode == SIMPLE_HEBBIAN:
            hidden_size=0
            # target_size=0
            separated = False
            clamp = SOFT_CLAMP
            continuous = False
            learning_function = Hebbian

        self.recurrent_size = input_size + (hidden_size or 0)
        if separated and target_size:
            self.recurrent_size += target_size
            self.target_start = input_size + (hidden_size or 0)
            self._target_included = True
        else:
            self.target_start = 0
            self._target_included = False
        self.target_end = self.target_start + target_size
        size = self.recurrent_size

        default_variable = [np.zeros(input_size), np.zeros(self.recurrent_size)]
        # Set InputPort sizes in _instantiate_input_ports,
        #    so that there is no conflict with parsing of Mechanism's size
        input_ports = [INPUT, RECURRENT]
        if self._target_included:
            default_variable.append(np.zeros(target_size))
            input_ports.append(TARGET)

        if additional_input_ports:
            if isinstance(additional_input_ports, list):
                input_ports += additional_input_ports
            else:
                input_ports.append(additional_input_ports)

        combination_function = combination_function or self.combination_function

        output_ports = [OUTPUT_ACTIVITY, CURRENT_ACTIVITY, ACTIVITY_DIFFERENCE]
        if additional_output_ports:
            if isinstance(additional_output_ports, list):
                output_ports += additional_output_ports
            else:
                output_ports.append(additional_output_ports)

        super().__init__(
            default_variable=default_variable,
            size=size,
            input_ports=input_ports,
            combination_function=combination_function,
            function=function,
            matrix=matrix,
            auto=auto,
            hetero=hetero,
            has_recurrent_input_port=True,
            integrator_function=integrator_function,
            initial_value=initial_value,
            noise=noise,
            integrator_mode=integrator_mode,
            integration_rate=integration_rate,
            clip=clip,
            enable_learning=enable_learning,
            learning_rate=learning_rate,
            learning_function=learning_function,
            learning_condition=CONVERGENCE,
            output_ports=output_ports,
            mode=mode,
            minus_phase_termination_condition=minus_phase_termination_condition,
            minus_phase_termination_threshold=minus_phase_termination_threshold,
            plus_phase_termination_condition=plus_phase_termination_condition,
            plus_phase_termination_threshold=plus_phase_termination_threshold,
            phase_convergence_function=phase_convergence_function,
            phase_convergence_threshold=minus_phase_termination_threshold,
            max_passes=max_passes,
            continuous=continuous,
            input_size=input_size,
            hidden_size=hidden_size,
            target_size=target_size,
            separated=separated,
            clamp=clamp,
            params=params,
            name=name,
            prefs=prefs,
            **kwargs
        )

    def _validate_params(self, request_set, target_set=None, context=None):

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        # Make sure that the size of the INPUT and TARGET InputPorts are <= size of RECURRENT InputPort
        if self._target_included and self.input_size != self.target_size:
            raise ContrastiveHebbianError("{} is {} for {} must equal {} ({}) must equal {} ({})} ".
                                          format(repr(SEPARATED), repr(True), self.name,
                                                 repr(INPUT_SIZE), self.input_size,
                                                 repr(TARGET_SIZE), self.target_size))

    def _instantiate_input_ports(self, input_ports=None, reference_value=None, context=None):

        # Assign InputPort specification dictionaries for required InputPorts
        sizes = dict(INPUT=self.input_size, RECURRENT=self.recurrent_size, TARGET=self.target_size)
        for i, input_port in enumerate((s for s in self.input_ports if s in {INPUT, TARGET, RECURRENT})):
            self.input_ports[i] = {NAME:input_port, SIZE: sizes[input_port]}

        super()._instantiate_input_ports(input_ports, reference_value, context)

        self.input_ports[RECURRENT].internal_only = True
        if self._target_included:
            self.input_ports[TARGET].internal_only = True

    def _instantiate_attributes_before_function(self, function=None, context=None):
        super()._instantiate_attributes_before_function(function=function, context=context)

        # Set minus_phase activity, plus_phase, current_activity and initial_value
        #    all  to zeros with size of Mechanism's array
        # Should be OK to use attributes here because initialization should only occur during None context
        self._set_multiple_parameter_values(
            context,
            initial_value=self.input_ports[RECURRENT].socket_template,
            current_activity=self.input_ports[RECURRENT].socket_template,
            minus_phase_activity=self.input_ports[RECURRENT].socket_template,
            plus_phase_activity=self.input_ports[RECURRENT].socket_template,
            execution_phase=None,
        )
        self.defaults.initial_value = copy.deepcopy(self.input_ports[RECURRENT].socket_template)
        self.defaults.current_activity = copy.deepcopy(self.input_ports[RECURRENT].socket_template)
        self.defaults.minus_phase_activity = copy.deepcopy(self.input_ports[RECURRENT].socket_template)
        self.defaults.plus_phase_activity = copy.deepcopy(self.input_ports[RECURRENT].socket_template)
        self.defaults.execution_phase = None

        if self._target_included:
            self.parameters.output_activity._set(self.input_ports[TARGET].socket_template, context)

    @tc.typecheck
    def _instantiate_recurrent_projection(self,
                                          mech: Mechanism,
                                          # this typecheck was failing, I didn't want to fix (7/19/17 CW)
                                          # matrix:is_matrix=HOLLOW_MATRIX,
                                          matrix=HOLLOW_MATRIX,
                                          context=None):
        """Instantiate an AutoAssociativeProjection from Mechanism to itself
        """
        if isinstance(matrix, str):
            size = len(mech.defaults.variable[0])
            matrix = get_matrix(matrix, size, size)

        return AutoAssociativeProjection(owner=mech,
                                         sender=self.output_ports[CURRENT_ACTIVITY],
                                         receiver=self.input_ports[RECURRENT],
                                         matrix=matrix,
                                         name=mech.name + ' recurrent projection')

    def _execute(self,
                 variable=None,
                 context=None,
                 function_variable=None,
                 runtime_params=None,
                 ):

        # Initialize execution_phase as minus_phase
        if self.parameters.execution_phase._get(context) is None:
            self.parameters.execution_phase._set(MINUS_PHASE, context)

        if self.parameters.execution_phase._get(context) == MINUS_PHASE:
            self.parameters.current_termination_threshold._set(
                    self.parameters.minus_phase_termination_threshold._get(context), context)
            self.parameters.current_termination_condition._set(self.minus_phase_termination_condition, context)
            self.parameters.phase_execution_count._set(0, context)

        if self.parameters.is_finished_flag._get(context):
        # if self.parameters.is_finished_._get(context):
            # If current execution follows completion of a previous trial,
            #    zero activity for input from recurrent projection so that
            #    input does not contain residual activity of previous trial
            variable[RECURRENT_INDEX] = self.input_ports[RECURRENT].socket_template

        # self.parameters.is_finished_._set(False, context)
        self.parameters.is_finished_flag._set(False, context)

        # Note _parse_function_variable selects actual input to function based on execution_phase
        current_activity = super()._execute(variable,
                                            context=context,
                                            runtime_params=runtime_params,
                                            )

        self.parameters.phase_execution_count._set(self.parameters.phase_execution_count._get(context) + 1, context)

        current_activity = np.squeeze(current_activity)
        # Set value of primary OutputPort to current activity
        self.parameters.current_activity._set(current_activity, context)

        # This is the first trial, so can't test for convergence
        #    (since that requires comparison with value from previous trial)
        # (here, the "previous value" is the current parameter value
        # because it has not been updated with the results of this
        # execution yet)
        if self.parameters.value._get(context) is None:
            return current_activity

        current_termination_condition = self.parameters.current_termination_condition._get(context)

        if current_termination_condition == CONVERGENCE:
            self.parameters.phase_convergence_threshold._set(
                    self.parameters.current_termination_threshold._get(context),context)
            self.parameters.phase_terminated._set(
                    self.is_converged(np.atleast_2d(current_activity), context), context)
        elif current_termination_condition == COUNT:
            self.parameters.phase_terminated._set(
                (self.parameters.phase_execution_count._get(context) ==
                 self.parameters.current_termination_threshold._get(context)),
                context)
        else:
            phase_str = repr('PLUS_PHASE') if self.parameters.execution_phase._get(context) == PLUS_PHASE \
                                           else repr('MINUS_PHASE')
            raise ContrastiveHebbianError(f"Unrecognized {repr('current_termination_condition')} specification "
                                          f"({current_termination_condition}) in execution of {phase_str} of "
                                          f"{self.name}.")

        if self.parameters.phase_terminated._get(context):
            # Terminate if this is the end of the plus phase, prepare for next trial
            if self.parameters.execution_phase._get(context) == PLUS_PHASE:
                # Store activity from last execution in plus phase
                self.parameters.plus_phase_activity._set(current_activity, context)
                # # Set value of primary outputPort to activity at end of plus phase
                self.parameters.current_activity._set(current_activity, context)
                self.parameters.is_finished_flag._set(True, context)

            # Otherwise, prepare for start of plus phase on next execution
            else:
                # Store activity from last execution in plus phase
                self.parameters.minus_phase_activity._set(self.parameters.current_activity._get(context), context)
                # Use initial_value attribute to initialize, for the minus phase,
                #    both the integrator_function's previous_value
                #    and the Mechanism's current activity (which is returned as its input)
                if not self.continuous and self.parameters.integrator_mode._get(context):
                    self.reset(self.parameters.initial_value._get(context), context=context)
                    self.parameters.current_activity._set(self.parameters.initial_value._get(context), context)
                self.parameters.current_termination_threshold._set(self.parameters.plus_phase_termination_threshold._get(context), context)
                self.parameters.current_termination_condition._set(self.parameters.plus_phase_termination_condition._get(context), context)

            # Switch execution_phase
            self.parameters.execution_phase._set(not self.parameters.execution_phase._get(context), context)
            self.parameters.phase_execution_count._set(0, context)

        return current_activity
        # return self.current_activity

    # TODO: remove this override after making combination_function its
    # own Component?
    def _parse_integrator_function_variable(self, variable, context=None):
        return self.combination_function(variable=variable, context=context)

    def _parse_phase_convergence_function_variable(self, variable):
        # determines shape only
        return np.asarray([variable[0], variable[0]])

    def combination_function(self, variable=None, context=None):
        # IMPLEMENTATION NOTE: use try and except here for efficiency: care more about execution than initialization
        # IMPLEMENTATION NOTE: separated vs. overlapping input and target handled by assignment of target_start in init

        MINUS_PHASE_INDEX = INPUT_INDEX
        PLUS_PHASE_INDEX = TARGET_INDEX

        try:  # Execution
            if self.parameters.execution_phase._get(context) == PLUS_PHASE:
                if self.clamp == HARD_CLAMP:
                    variable[RECURRENT_INDEX][:self.input_size] = variable[MINUS_PHASE_INDEX]
                    if self.mode == SIMPLE_HEBBIAN:
                        return variable[RECURRENT_INDEX]
                    else:
                        variable[RECURRENT_INDEX][self.target_start:self.target_end] = variable[PLUS_PHASE_INDEX]
                else:
                    variable[RECURRENT_INDEX][:self.input_size] += variable[MINUS_PHASE_INDEX]
                    if self.mode == SIMPLE_HEBBIAN:
                        return variable[RECURRENT_INDEX]
                    else:
                        variable[RECURRENT_INDEX][self.target_start:self.target_end] += variable[PLUS_PHASE_INDEX]
            else:
                if self.mode == SIMPLE_HEBBIAN:
                    return variable[RECURRENT_INDEX]
                if self.clamp == HARD_CLAMP:
                    variable[RECURRENT_INDEX][:self.input_size] = variable[MINUS_PHASE_INDEX]
                else:
                    variable[RECURRENT_INDEX][:self.input_size] += variable[MINUS_PHASE_INDEX]
        except:  # Initialization
            pass

        return variable[RECURRENT_INDEX]

    def delta(self, value=NotImplemented, context=None):
        if value is NotImplemented:
            self.phase_convergence_function([
                self.parameters.value._get(context)[0],
                self.parameters.value.get_previous(context)[0]
            ])
        else:
            return self.phase_convergence_function([value[0], self.parameters.value._get(context)[0]])

    @handle_external_context()
    def is_converged(self, value=NotImplemented, context=None):
        phase_convergence_threshold = self.parameters.phase_convergence_threshold._get(context)
        # Check for convergence
        if (
            phase_convergence_threshold is not None
            and self.parameters.value.get_previous(context) is not None
            and self.initialization_status != ContextFlags.INITIALIZING
        ):
            if self.delta(value, context) <= phase_convergence_threshold:
                return True
            elif self.get_current_execution_time(context).pass_ >= self.max_passes:
                phase_str = repr('PLUS_PHASE') if self.parameters.execution_phase._get(context) == PLUS_PHASE \
                    else repr('MINUS_PHASE')
                raise ContrastiveHebbianError(f"Maximum number of executions ({self.max_passes}) has occurred "
                                              f"before reaching convergence_threshold "
                                              f"({phase_convergence_threshold}) for {self.name} in "
                                              f"{phase_str} of trial {self.get_current_execution_time(context).trial} "
                                              f"of run {self.get_current_execution_time(context).run}.")
            else:
                return False
        # Otherwise just return True
        else:
            return None

    @property
    def _learning_signal_source(self):
        """Override default to use ACTIVITY_DIFFERENCE as source of learning signal
        """
        return self.output_ports[ACTIVITY_DIFFERENCE]

    @property
    def input_activity(self):
        return self.current_activity[:self.input_size]

    @property
    def hidden_activity(self):
        if self.hidden_size:
            return self.current_activity[self.input_size:self.target_start]

    @property
    def target_activity(self):
        if self.target_size:
            return self.current_activity[self.target_start:self.target_end]

    @property
    def recurrent_activity(self):
        return self.current_activity

    @property
    def recurrent_size(self):
        return self._recurrent_size

    @recurrent_size.setter
    def recurrent_size(self, value):
        self._recurrent_size = value



import numpy as np
import pytest

from psyneulink.core.components.functions.transferfunctions import Linear, Logistic
from psyneulink.core.components.functions.userdefinedfunction import UserDefinedFunction
from psyneulink.core.components.mechanisms.processing import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing import TransferMechanism
from psyneulink.core.compositions.composition import Composition

import psyneulink.core.llvm as pnlvm

class TestBinaryOperations:
    @pytest.mark.parametrize("param1, param2", [
                        (1, 2),
                        (np.ones(2), 2),
                        (2, np.ones(2)),
                        (np.ones((2, 2)), 2),
                        (2, np.ones((2, 2))),
                        (np.ones(2), np.array([1, 2])),
                        (np.ones((2, 2)), np.array([[1, 2], [3, 4]])),
                        ], ids=["scalar-scalar", "vec-scalar", "scalar-vec", "mat-scalar", "scalar-mat", "vec-vec", "mat-mat"])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_add(self, param1, param2, bin_execute, benchmark):
        # default val is same shape as expected output
        def myFunction(_, param1, param2):
            # we only use param1 and param2 to avoid automatic shape changes of the variable
            return param1 + param2

        U = UserDefinedFunction(custom_function=myFunction, param1=param1, param2=param2)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, 0)
        assert np.allclose(val, param1 + param2)

    @pytest.mark.parametrize("param1, param2", [
                        (1, 2),
                        (np.ones(2), 2),
                        (2, np.ones(2)),
                        (np.ones((2, 2)), 2),
                        (2, np.ones((2, 2))),
                        (np.ones(2), np.array([1, 2])),
                        (np.ones(2), np.array([2.])),
                        (np.ones((2, 2)), np.array([[1, 2], [3, 4]])),
                        ], ids=["scalar-scalar", "vec-scalar", "scalar-vec", "mat-scalar", "scalar-mat", "vec-vec", "vec-vec-differing", "mat-mat"])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_mul(self, param1, param2, bin_execute, benchmark):
        # default val is same shape as expected output
        def myFunction(_, param1, param2):
            # we only use param1 and param2 to avoid automatic shape changes of the variable
            return param1 * param2

        U = UserDefinedFunction(custom_function=myFunction, param1=param1, param2=param2)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, 0)
        assert np.allclose(val, param1 * param2)

    @pytest.mark.parametrize("param1, param2", [
                    (1, 2),
                    (np.ones(2), 2),
                    (2, np.ones(2)),
                    (np.ones((2, 2)), 2),
                    (2, np.ones((2, 2))),
                    (np.ones(2), np.array([1, 2])),
                    (np.ones(2), np.array([2.])),
                    (np.ones((2, 2)), np.array([[1, 2], [3, 4]])),
                    ], ids=["scalar-scalar", "vec-scalar", "scalar-vec", "mat-scalar", "scalar-mat", "vec-vec", "vec-vec-differing", "mat-mat"])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_div(self, param1, param2, bin_execute, benchmark):
        # default val is same shape as expected output
        def myFunction(_, param1, param2):
            # we only use param1 and param2 to avoid automatic shape changes of the variable
            return param1 / param2

        U = UserDefinedFunction(custom_function=myFunction, param1=param1, param2=param2)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, 0)
        assert np.allclose(val, np.divide(param1, param2))

    @pytest.mark.parametrize("op", [ # parameter is string since compiled udf doesn't support closures as of present
                        "AND",
                        "OR",
                        ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_boolop(self, op, bin_execute, benchmark):
        if op == "AND":
            def myFunction(variable):
                var1 = True
                var2 = False
                # compiled UDFs don't support python bool type outputs
                if var1 and var2:
                    return 0.0
                else:
                    return 1.0
        elif op == "OR":
            def myFunction(variable):
                var1 = True
                var2 = False
                # compiled UDFs don't support python bool type outputs
                if var1 or var2:
                    return 1.0
                else:
                    return 0.0

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[0])
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [0])
        assert val == 1.0

    @pytest.mark.parametrize("op,var1,var2,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                        ("Eq", 1.0, 2.0, 0.0),
                        ("NotEq", 1.0, 2.0, 1.0),
                        ("Lt", 1.0, 2.0, 1.0),
                        ("LtE", 1.0, 2.0, 1.0),
                        ("Gt", 1.0, 2.0, 0.0),
                        ("GtE", 1.0, 2.0, 0.0),
                        ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_cmpop(self, op, var1, var2, expected, bin_execute, benchmark):
        # we explicitly use np here to ensure that the result is castable to float in the scalar-scalar case
        if op == "Eq":
            def myFunction(variable, var1, var2):
                if var1 == var2:
                    return 1.0
                else:
                    return 0.0
        elif op == "NotEq":
            def myFunction(variable, var1, var2):
                if var1 != var2:
                    return 1.0
                else:
                    return 0.0
        elif op == "Lt":
            def myFunction(variable, var1, var2):
                if var1 < var2:
                    return 1.0
                else:
                    return 0.0
        elif op == "LtE":
            def myFunction(variable, var1, var2):
                if var1 <= var2:
                    return 1.0
                else:
                    return 0.0
        elif op == "Gt":
            def myFunction(variable, var1, var2):
                if var1 > var2:
                    return 1.0
                else:
                    return 0.0
        elif op == "GtE":
            def myFunction(variable, var1, var2):
                if var1 >= var2:
                    return 1.0
                else:
                    return 0.0

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[0], var1=var1, var2=var2)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [0])
        assert np.allclose(expected, val)

    @pytest.mark.parametrize("op,var1,var2,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                        ("Eq", 1.0, 2.0, 0.0),
                        ("Eq", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                        ("Eq", 1.0, [1.0, 2.0], [1.0, 0.0]),
                        ("Eq", [2.0, 1.0], 1.0, [0.0, 1.0]),
                        ("Eq", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 0.0], [0.0, 0.0]]),
                        ("Eq", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 0.0], [0.0, 0.0]]),
                        ("Eq", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                        ("NotEq", 1.0, 2.0, 1.0),
                        ("NotEq", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                        ("NotEq", 1.0, [1.0, 2.0], [0.0, 1.0]),
                        ("NotEq", [2.0, 1.0], 1.0, [1.0, 0.0]),
                        ("NotEq", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 1.0], [1.0, 1.0]]),
                        ("NotEq", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 1.0], [1.0, 1.0]]),
                        ("NotEq", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                        ("Lt", 1.0, 2.0, 1.0),
                        ("Lt", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                        ("Lt", 1.0, [1.0, 2.0], [0.0, 1.0]),
                        ("Lt", [2.0, 1.0], 1.0, [0.0, 0.0]),
                        ("Lt", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 0.0], [0.0, 0.0]]),
                        ("Lt", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 1.0], [1.0, 1.0]]),
                        ("Lt", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                        ("LtE", 1.0, 2.0, 1.0),
                        ("LtE", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                        ("LtE", 1.0, [1.0, 2.0], [1.0, 1.0]),
                        ("LtE", [2.0, 1.0], 1.0, [0.0, 1.0]),
                        ("LtE", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 0.0], [0.0, 0.0]]),
                        ("LtE", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                        ("LtE", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                        ("Gt", 1.0, 2.0, 0.0),
                        ("Gt", [1.0, 2.0], [1.0, 2.0], [0.0, 0.0]),
                        ("Gt", 1.0, [1.0, 2.0], [0.0, 0.0]),
                        ("Gt", [2.0, 1.0], 1.0, [1.0, 0.0]),
                        ("Gt", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[0.0, 1.0], [1.0, 1.0]]),
                        ("Gt", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                        ("Gt", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[0.0, 0.0], [0.0, 0.0]]),
                        ("GtE", 1.0, 2.0, 0.0),
                        ("GtE", [1.0, 2.0], [1.0, 2.0], [1.0, 1.0]),
                        ("GtE", 1.0, [1.0, 2.0], [1.0, 0.0]),
                        ("GtE", [2.0, 1.0], 1.0, [1.0, 1.0]),
                        ("GtE", [[1.0, 2.0], [3.0, 4.0]], 1.0, [[1.0, 1.0], [1.0, 1.0]]),
                        ("GtE", 1.0, [[1.0, 2.0], [3.0, 4.0]], [[1.0, 0.0], [0.0, 0.0]]),
                        ("GtE", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]], [[1.0, 1.0], [1.0, 1.0]]),
                        ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_cmpop_numpy(self, op, var1, var2, expected, bin_execute, benchmark):
        # we explicitly use np here to ensure that the result is castable to float in the scalar-scalar case
        if op == "Eq":
            def myFunction(variable, var1, var2):
                return np.equal(var1, var2).astype(float)
        elif op == "NotEq":
            def myFunction(variable, var1, var2):
                return np.not_equal(var1, var2).astype(float)
        elif op == "Lt":
            def myFunction(variable, var1, var2):
                return np.less(var1, var2).astype(float)
        elif op == "LtE":
            def myFunction(variable, var1, var2):
                return np.less_equal(var1, var2).astype(float)
        elif op == "Gt":
            def myFunction(variable, var1, var2):
                return np.greater(var1, var2).astype(float)
        elif op == "GtE":
            def myFunction(variable, var1, var2):
                return np.greater_equal(var1, var2).astype(float)

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[0], var1=var1, var2=var2)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [0])
        assert np.allclose(expected, val)

class TestUserDefFunc:

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func(self, bin_execute, benchmark):
        def myFunction(variable, param1, param2):
            return variable * 2 + param2

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0]], param2=3)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [1, 3])
        assert np.allclose(val, [[5, 9]])

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_branching(self, bin_execute, benchmark):
        def myFunction(variable, param1, param2):
            if variable[0][0] > 0 and variable[0][1] > 0:
                return variable * 2 + param2
            else:
                return variable * -2 + param2

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0]], param2=3)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [[1, 3]])
        assert np.allclose(val, [[5, 9]])
        val2 = e([[-1, 3]])
        assert np.allclose(val2, [[5, -3]])

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_variable_index(self, bin_execute, benchmark):
        def myFunction(variable):
            variable[0][0] = variable[0][0] + 5
            variable[0][1] = variable[0][1] + 7
            return variable

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0]])
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, [[1, 3]])
        assert np.allclose(val, [[6, 10]])

    @pytest.mark.parametrize("variable", [
                        (1),
                        (np.ones((2))),
                        (np.ones((2, 2)))
                        ], ids=["scalar", "vec-2d", "mat"])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_usub(self, variable, bin_execute, benchmark):
        def myFunction(variable, param):
            return -param

        U = UserDefinedFunction(custom_function=myFunction, default_variable=variable, param=variable)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, variable)
        assert np.allclose(val, -variable)

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_reward_func(self, bin_execute, benchmark):
        variable = [[1,2,3,4]]
        def myFunction(x,t0=0.48):
            return (x[0][0]>0).astype(float) * (x[0][2]>0).astype(float) / (np.max([x[0][1],x[0][3]]) + t0)
        U = UserDefinedFunction(custom_function=myFunction, default_variable=variable, param=variable)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, variable)
        assert np.allclose(val, 0.2232142857142857)

    @pytest.mark.parametrize("dtype, expected", [ # parameter is string since compiled udf doesn't support closures as of present
                        ("SCALAR", 1.0),
                        ("VECTOR", [1,2]),
                        ("MATRIX", [[1,2],[3,4]]),
                        ("BOOL", 1.0),
                        ("TUPLE", (1, 2, 3, 4))
                        ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_assign(self, dtype, expected, bin_execute, benchmark):
        if dtype == "SCALAR":
            def myFunction(variable):
                var = 1.0
                return var
        elif dtype == "VECTOR":
            def myFunction(variable):
                var = [1,2]
                return var
        elif dtype == "MATRIX":
            def myFunction(variable):
                var = [[1,2],[3,4]]
                return var
        elif dtype == "BOOL":
            def myFunction(variable):
                var = True
                return 1.0
        elif dtype == "TUPLE":
            def myFunction(variable):
                var = (1, 2, 3, 4)
                return var

        U = UserDefinedFunction(custom_function=myFunction, default_variable=0)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, 0)
        assert np.allclose(val, expected)

    @pytest.mark.parametrize("op,variable,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                    ("TANH", [[1, 3]], [0.76159416, 0.99505475]),
                    ("EXP", [[1, 3]], [2.71828183, 20.08553692]),
                    ("SHAPE", [1, 2], [2]),
                    ("SHAPE", [[1, 3]], [1, 2]),
                    ("ASTYPE_FLOAT", [1], [1.0]),
                    ("ASTYPE_INT", [-1.5], [-1.0]),
                    ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_numpy(self, op, variable, expected, bin_execute, benchmark):
        if op == "TANH":
            def myFunction(variable):
                return np.tanh(variable)
        elif op == "EXP":
            def myFunction(variable):
                return np.exp(variable)
        elif op == "SHAPE":
            def myFunction(variable):
                return variable.shape
        elif op == "ASTYPE_FLOAT":
            def myFunction(variable):
                return variable.astype(float)
        elif op == "ASTYPE_INT":
            # return types cannot be integers, so we cast back to float and check for truncation
            def myFunction(variable):
                return variable.astype(int).astype(float)

        U = UserDefinedFunction(custom_function=myFunction, default_variable=variable)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, variable)
        assert np.allclose(val, expected)

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="UDF in Mechanism")
    def test_udf_in_mechanism(self, bin_execute, benchmark):
        def myFunction(variable, param1, param2):
            return sum(variable[0]) + 2

        myMech = ProcessingMechanism(function=myFunction, size=4, name='myMech')
        # assert 'param1' in myMech.parameter_ports.names # <- FIX reinstate when problem with function params is fixed
        # assert 'param2' in myMech.parameter_ports.names # <- FIX reinstate when problem with function params is fixed
        if bin_execute == 'LLVM':
            e = pnlvm.execution.MechExecution(myMech).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.MechExecution(myMech).cuda_execute
        else:
            e = myMech.execute
        val = benchmark(e, [-1, 2, 3, 4])
        assert np.allclose(val, [[10]])

    @pytest.mark.parametrize("op,variable,expected", [ # parameter is string since compiled udf doesn't support closures as of present
                    ("SUM", [1.0, 3.0], 4),
                    ("SUM", [[1.0], [3.0]], [4.0]),
                    ("LEN", [1.0, 3.0], 2),
                    ("LEN", [[1.0], [3.0]], 2),
                    ("LEN_TUPLE", [0, 0], 2),
                    ("MAX", [0.0, 0.0], 0),
                    ("MAX", [1.0, 2.0], 2),
                    ("MAX", [[2.0, 1.0], [6.0, 2.0]], 6),
                    ])
    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="Function UDF")
    def test_user_def_func_builtin(self, op, variable, expected, bin_execute, benchmark):
        if op == "SUM":
            def myFunction(variable):
                return sum(variable)
        elif op == "LEN":
            def myFunction(variable):
                return len(variable)
        elif op == "LEN_TUPLE":
            def myFunction(variable):
                return len((1,2))
        elif op == "MAX":
            def myFunction(variable):
                return np.max(variable)

        U = UserDefinedFunction(custom_function=myFunction, default_variable=variable)
        if bin_execute == 'LLVM':
            e = pnlvm.execution.FuncExecution(U).execute
        elif bin_execute == 'PTX':
            e = pnlvm.execution.FuncExecution(U).cuda_execute
        else:
            e = U
        val = benchmark(e, variable)
        assert np.allclose(val, expected)

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('LLVMExec', marks=pytest.mark.llvm),
                                             pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                             pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                             pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="UDF as Composition Origin")
    def test_udf_composition_origin(self, bin_execute, benchmark):
        def myFunction(variable, context):
            return [variable[0][1], variable[0][0]]

        myMech = ProcessingMechanism(function=myFunction, size=3, name='myMech')
        T = TransferMechanism(size=2, function=Linear)
        c = Composition(pathways=[myMech, T])
        benchmark(c.run, inputs={myMech: [[1, 3, 5]]}, bin_execute=bin_execute)
        assert np.allclose(c.results[0][0], [3, 1])

    @pytest.mark.parametrize("bin_execute", ['Python',
                                             pytest.param('LLVM', marks=pytest.mark.llvm),
                                             pytest.param('LLVMExec', marks=pytest.mark.llvm),
                                             pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                             pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                             pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                            ])
    @pytest.mark.benchmark(group="UDF as Composition Terminal")
    def test_udf_composition_terminal(self, bin_execute, benchmark):
        def myFunction(variable, context):
            return [variable[0][2], variable[0][0]]

        myMech = ProcessingMechanism(function=myFunction, size=3, name='myMech')
        T2 = TransferMechanism(size=3, function=Linear)
        c2 = Composition(pathways=[[T2, myMech]])
        benchmark(c2.run, inputs={T2: [[1, 2, 3]]}, bin_execute=bin_execute)
        assert(np.allclose(c2.results[0][0], [3, 1]))

    def test_udf_with_pnl_func(self):
        L = Logistic(gain=2)

        def myFunction(variable, context):
            return L(variable) + 2

        U = UserDefinedFunction(custom_function=myFunction, default_variable=[[0, 0, 0]])
        myMech = ProcessingMechanism(function=myFunction, size=3, name='myMech')
        val1 = myMech.execute(input=[1, 2, 3])
        val2 = U.execute(variable=[[1, 2, 3]])
        assert np.allclose(val1, val2)
        assert np.allclose(val1, L([1, 2, 3]) + 2)

    def test_udf_creates_parameter_ports(self):
        def func(input=[[0], [0]], p=0, q=1):
            return (p + q) * input

        m = ProcessingMechanism(
            default_variable=[[0], [0]],
            function=UserDefinedFunction(func)
        )

        assert len(m.parameter_ports) == 2
        assert 'p' in m.parameter_ports.names
        assert 'q' in m.parameter_ports.names

    @pytest.fixture
    def mech_with_autogenerated_udf(self):
        def func(input=[[0], [0]], p=0, q=1):
            return (p + q) * input

        m = ProcessingMechanism(
            default_variable=[[0], [0]],
            function=func
        )

        return m

    def test_mech_autogenerated_udf_execute(self, mech_with_autogenerated_udf):
        # Test if execute is working with auto-defined udf's
        val1 = mech_with_autogenerated_udf.execute(input=[[1], [1]])
        assert np.allclose(val1, np.array([[1], [1]]))

    def test_autogenerated_udf(self, mech_with_autogenerated_udf):
        assert isinstance(mech_with_autogenerated_udf.function, UserDefinedFunction)

    def test_autogenerated_udf_creates_parameter_ports(self, mech_with_autogenerated_udf):
        assert len(mech_with_autogenerated_udf.parameter_ports) == 2
        assert 'p' in mech_with_autogenerated_udf.parameter_ports.names
        assert 'q' in mech_with_autogenerated_udf.parameter_ports.names

    def test_autogenerated_udf_creates_parameters(self, mech_with_autogenerated_udf):
        assert hasattr(mech_with_autogenerated_udf.function.parameters, 'p')
        assert hasattr(mech_with_autogenerated_udf.function.parameters, 'q')

        assert mech_with_autogenerated_udf.function.parameters.p.default_value == 0
        assert mech_with_autogenerated_udf.function.parameters.q.default_value == 1

        assert not hasattr(mech_with_autogenerated_udf.function.class_parameters, 'p')
        assert not hasattr(mech_with_autogenerated_udf.function.class_parameters, 'q')

    def test_autogenerated_udf_parameters_states_have_source(self, mech_with_autogenerated_udf):
        for p in mech_with_autogenerated_udf.parameter_ports:
            assert p.source is getattr(mech_with_autogenerated_udf.function.parameters, p.name)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  GatingMechanism ************************************************

"""

Contents
--------

  * `GatingMechanism_Overview`
  * `GatingMechanism_Creation`
  * `GatingMechanism_Structure`
      - `GatingMechanism_Input`
      - `GatingMechanism_Function`
  * `GatingMechanism_Execution`
  * `GatingMechanism_Class_Reference`


.. _GatingMechanism_Overview:

Overview
--------

A GatingMechanism is a subclass of `ControlMechanism` that is restricted to using only `GatingSignals <GatingSignal>`,
which modulate the `input <Mechanism_InputPorts>` or `output <Mechanism_InputPorts>` of a `Mechanism <Mechanism>`,
but not the paramaters of its `function <Mechanism_Base.function>`.  Accordingly, its constructor has a **gate**
argument in place of a **control** argument.  It also lacks several attributes related to control, including those
related to costs and net_outcome.  In all other respects it is identical to its parent class, ControlMechanism.

.. _GatingMechanism_Creation:

Creating A GatingMechanism
---------------------------

A GatingMechanism is created by calling its constructor.  When a GatingMechanism is created, the OutputPorts it
monitors and the `InputPorts <InputPort>` and/or `OutputPorts <OutputPort>` it modulates can be specified in the
**montior_for_gating** and **gate** arguments of its constructor, respectively.  Each can be specified in several
ways, paralleling those used for a ControlMechanism, and described in `ControlMechanism_Monitor_for_Control` and
`ControlMechanism_ControlSignals` respectively. If neither the **montior_for_gating** or **gate** arguments is
specified, then only the GatingMechanism is constructed, and its inputs and the InputPorts and/or OutputPorts it
modulates must be specified in some other way.
COMMENT:
TBI FOR COMPOSITION
A GatingMechanism is also created automatically if `gating
is specified <GatingMechanism_Specifying_Gating>` for an `InputPort`, `OutputPort` or `Mechanism <Mechanism>`,
in which case a `GatingProjection` is automatically created that projects from the GatingMechanism to the specified
target.
COMMENT

.. _GatingMechanism_Specifying_Gating:

*Specifying gating*
~~~~~~~~~~~~~~~~~~~

A GatingMechanism is used to modulate the value of an `InputPort` or `OutputPort`. An InputPort or OutputPort can
be specified for gating by assigning it a `GatingProjection` or `GatingSignal` anywhere that the Projections to a Port
or its `ModulatorySignals can be specified <State_Creation>`.  A `Mechanism <Mechanism>` can also be specified for
gating, in which case the `primary InputPort <InputPort_Primary>` of the specified Mechanism is used.  Ports
(and/or Mechanisms) can also be specified in the  **gate** argument of the constructor for a GatingMechanism. The
**gate** argument must be a list, each item of which must refer to one or more Ports (or the Mechanism(s) to which
they belong) to be gated by that GatingSignal.  The specification for each item in the list can use any of the forms
used to `specify a GatingSignal <GatingSignal_Specification>`.

.. _GatingMechanism_GatingSignals:

GatingSignals
^^^^^^^^^^^^^

A `GatingSignal` is created for each item listed in the **gate** argument of the constructor, and all of the
GatingSignals for a GatingMechanism are listed in its `gating_signals <GatingMechanism.gating_signals>` attribute.
Each GatingSignal is assigned one or more `GatingProjections <GatingProjection>` to the InputPort(s) and/or
OutputPort(s) it gates. By default, the `function <GatingMechanism.function>` of GatingMechanism generates a `value
<GatingMechanism.value>` -- its `gating_allocation <GatingSignal.gating_allocation>` -- with a single item, that is
used by all of the GatingMechanism's GatingSignals.  However,  if a custom `function <GatingMechanism.function>` is
specified that generates a `gating_allocation <GatingSignal.gating_allocation>` with more than one item, different
GatingSignals can be assigned to the different items (see `GatingMechanism_Function` below).

.. _GatingMechanism_Modulation:

Modulation
^^^^^^^^^^

Each GatingMechanism has a `modulation <GatingSignal.modulation>` attribute, that provides a default for the way
in which its GatingSignals modulate the value of the Ports they gate (see `modulation <ModulatorySignal_Modulation>`
for an explanation of how this attribute is specified and used to modulate the value of a Port).  Each GatingSignal
uses this value, unless its value is `individually specified <GatingSignal_Modulation>`.

.. _GatingMechanism_Structure:

Structure
---------

.. _GatingMechanism_Input:

*Input*
~~~~~~~

The input to a GatingMechanism is determined in the same manner as the `input <ControlMechanism_Input>` to
any `ControlMechanism`.

.. _GatingMechanism_Function:

*Function*
~~~~~~~~~~

A GatingMechanism's `function <GatingMechanism.function>` is determined and operates in the same manner as the
`function <ControlMechanism_Function>` of any `ControlMechanism`.

.. _GatingMechanism_Output:

*Output*
~~~~~~~~

The OutputPorts of a GatingMechanism are `GatingSignals <GatingSignal>` (listed in its `gating_signals
<GatingMechanism.gating_signals>` attribute). It  has a `GatingSignal` for each `InputPort` and/or `OutputPort`
specified in the **gate** argument of its constructor, that sends a `GatingProjection` to those Ports.
The GatingSignals are listed in the `gating_signals <GatingMechanism.gating_signals>` attribute;  since they are a
type of `OutputPort`, they are also listed in the GatingMechanism's `output_ports <GatingMechanism.output_ports>`
attribute. The InputPorts and/or OutputPorts modulated by a GatingMechanism's GatingSignals can be displayed using
its `show <GatingMechanism.show>` method. If the GatingMechanism's `function <GatingMechanism.function>` generates a
`gating_allocation <GatingMechanism.gating_allocation>` with a single value (the default), then this is used as the
`allocation <GatingSignal.alloction>` for all of the GatingMechanism's `gating_signals
<GatingMechanism.gating_signals>`.  If the `gating_allocation <GatingMechanism.gating_allocation>` has multiple
items, and this is the same as the number of GatingSignals, then each GatingSignal is assigned the value of the
corresponding item in the `gating_allocation <GatingMechanism.gating_allocation>`.  If there is a different number of
`gating_signals <GatingMechanism.gating_signals>` than the number of items in the `gating_allocation
<GatingMechanism.gating_allocation>`, then the `index <GatingSignal.index>` attribute of each GatingSignal must be
specified (e.g., in a `specification dictionary <GatingSignal_Specification>` in the **gating_signal** argument of
the GatingMechanism's constructor), or an error is generated.  The `default_allocation
<GatingMechanism.default_allocation>` attribute can be used to specify a  default allocation for GatingSignals that
have not been assigned their own `default_allocation  <GatingSignal.default_allocation>`. The `allocation
<GatingSignal.allocation>` is used by each GatingSignal to determine its `intensity  <GatingSignal.intensity>`,
which is then assigned to the `value <GatingProjection.value>` of the GatingSignal's `GatingProjection`.   The `value
<GatingProjection.value>` of the GatingProjection is used to modify the value of the InputPort and/or OutputPort it
gates (see `GatingSignal_Modulation` for description of how a GatingSignal modulates the value of a parameter).

.. _GatingMechanism_Execution:

Execution
---------

COMMENT:
    FIX 5/8/20: REWORK TO FOLLOW CONTROLMECHANISM
COMMENT

A GatingMechanism executes in the same way as a `ProcessingMechanism <ProcessingMechanism>`, based on its place in the
Composition's `graph <Composition.graph>`.  Because `GatingProjections <GatingProjection>` are likely to introduce
cycles (recurrent connection loops) in the graph, the effects of a GatingMechanism and its projections will generally
not be applied in the first `TRIAL <TimeScale.TRIAL>` (see
COMMENT:
`Composition_Cycles_and_Feedback` and
COMMENT
**feedback** argument for the `add_projection <Composition.add_projection>`
method of `Composition` for a description of how to configure the initialization of feedback loops in a Composition;
also see `Scheduler` for a description of detailed ways in which a GatingMechanism and its dependents can be scheduled
to execute).

When executed, a GatingMechanism  uses its input to determine the value of its `gating_allocation
<GatingMechanism.gating_allocation>`, each item of which is used by a corresponding `GatingSignal` to determine its
`gating_signal <GatingSignal.gating_signal>` and assign to its `GatingProjections <GatingProjection>`. In the
subsequent `TRIAL <TimeScale.TRIAL>`, each GatingProjection's value is used by the Port to which it projects to
modulate the `value <Port_Base.value>` of that Port (see `modulation <ModulatorySignal_Modulation>` fon an explanation
of how the value of a Port is modulated).

.. note::
   A Port that receives a `GatingProjection` does not update its `value <Port_Base.value>` (and therefore does not
   reflect the influence of its `GatingSignal`) until that Port's owner Mechanism executes (see `Lazy Evaluation
   <Component_Lazy_Updating>` for an explanation of "lazy" updating).

.. _GatingMechanism_Class_Reference:

Class Reference
---------------

"""

import numpy as np
import typecheck as tc

from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.ports.modulatorysignals.gatingsignal import GatingSignal
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.defaults import defaultGatingAllocation
from psyneulink.core.globals.keywords import \
    GATING, GATING_PROJECTION, GATING_PROJECTIONS,GATING_SIGNAL,GATING_SIGNALS,GATING_SIGNAL_SPECS, \
    INIT_EXECUTE_METHOD_ONLY, MONITOR_FOR_CONTROL, MULTIPLICATIVE, PROJECTION_TYPE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities import ContentAddressableList, convert_to_list

__all__ = [
    'GatingMechanism', 'GatingMechanismError', 'GatingMechanismRegistry'
]

GatingMechanismRegistry = {}


def _is_gating_spec(spec):
    from psyneulink.core.components.projections.modulatory.gatingprojection import GatingProjection
    if isinstance(spec, tuple):
        return any(_is_gating_spec(item) for item in spec)
    if isinstance(spec, dict) and PROJECTION_TYPE in spec:
        return _is_gating_spec(spec[PROJECTION_TYPE])
    elif isinstance(spec, (GatingMechanism,
                           GatingSignal,
                           GatingProjection,
                           ControlMechanism)):
        return True
    elif isinstance(spec, type) and issubclass(spec, (GatingSignal,
                                                      GatingProjection,
                                                      GatingMechanism,
                                                      ControlMechanism)):
        return True
    elif isinstance(spec, str) and spec in {GATING, GATING_PROJECTION, GATING_SIGNAL}:
        return True
    else:
        return False


class GatingMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

def _gating_allocation_getter(owning_component=None, context=None):
    return owning_component.control_allocation

def _gating_allocation_setter(value, owning_component=None, context=None):
    owning_component.parameters.control_allocation._set(np.array(value), context)
    return value

# def _control_allocation_getter(owning_component=None, context=None):
#     from psyneulink.core.components.mechanisms.modulatory.controlmechanism import ControlMechanism
#     from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
#     raise GatingMechanismError(f"'control_allocation' attribute is not implemented on {owning_component.name};  "
#                                 f"consider using a {ControlMechanism.__name__} instead, "
#                                 f"or a {ControlMechanism.__name__} if both {ControlSignal.__name__}s and "
#                                 f"{GatingSignal.__name__}s are needed.")
#
# def _control_allocation_setter(value, owning_component=None, context=None, **kwargs):
#     from psyneulink.core.components.mechanisms.modulatory.controlmechanism import ControlMechanism
#     from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
#     raise GatingMechanismError(f"'control_allocation' attribute is not implemented on {owning_component.name};  "
#                                 f"consider using a {ControlMechanism.__name__} instead, "
#                                 f"or a {ControlMechanism.__name__} if both {ControlSignal.__name__}s and "
#                                 f"{GatingSignal.__name__}s are needed.")


class GatingMechanism(ControlMechanism):
    """
    GatingMechanism(                           \
        default_gating_allocation=None,        \
        monitor_for_gating=None,               \
        function=Linear(slope=1, intercept=0), \
        default_allocation=None,               \
        gate=None,                             \
        modulation=MULTIPLICATIVE)

    Subclass of `ModulatoryMechanism <ModulatoryMechanism>` that gates (modulates) the value(s) of one or more `Ports
    <Port>`.  See `Mechanism <Mechanism_Class_Reference>` for additional arguments and attributes.

    COMMENT:
        Description:
            # VERIFY:
            Protocol for instantiating unassigned GatingProjections (i.e., w/o a sender specified):
               If sender is not specified for a GatingProjection (e.g., in an InputPort or OutputPort tuple spec)
                   it is flagged for deferred_init() in its __init__ method
    COMMENT

    Arguments
    ---------

    default_gating_allocation : value, list or ndarray : default `defaultGatingAllocation`
        the default value for each of the GatingMechanism's GatingSignals;
        its length must equal the number of items specified in the **gate** argument.

    size : int, list or 1d np.array of ints
        specifies default_gating_allocation as an array of zeros if **default_gating_allocation** is not passed as an
        argument;  if **default_gating_allocation** is specified, it takes precedence over the specification of **size**.
        As an example, the following mechanisms are equivalent::
            T1 = TransferMechanism(size = [3, 2])
            T2 = TransferMechanism(default_variable = [[0, 0, 0], [0, 0]])

    monitor_for_gating : List[OutputPort or Mechanism] : default None
        specifies the `OutputPorts <OutputPort>` to be monitored by the `ObjectiveMechanism`, if specified in an
        **objective_mechanism** argument (see `ControlMechanism_ObjectiveMechanism`), or directly by the
        GatingMechanism itself if **objective_mechanism** is not specified.  If any specification is a Mechanism
        (rather than its OutputPort), its `primary OutputPort <OutputPort_Primary>` is used (see
        `ControlMechanism_Monitor_for_Control` for additional details).

    function : TransferFunction : default Linear(slope=1, intercept=0)
        specifies the function used to transform the GatingMechanism's `variable <GatingMechanism.variable>`
        to a `gating_allocation`.

    default_allocation : number, list or 1d array : None
        specifies the default_allocation of any `gating_signals <GatingMechanism.gating.signals>` for
        which the **default_allocation** was not specified in its constructor (see default_allocation
        <GatingMechanism.default_allocation>` for additional details).

    gate : list[GatingSignal, InputPort, OutputPort, Mechanism, tuple[str, Mechanism], or dict]
        specifies the `InputPorts <InputPort>` and/or `OutputPorts <OutputPorts>` to be gated by the
        GatingMechanism; the number of items must equal the length of the **default_gating_allocation**
        argument; if a `Mechanism <Mechanism>` is specified, its `primary InputPort <InputPort_Primary>`
        is used (see `GatingMechanism_GatingSignals for details).

    modulation : str : MULTIPLICATIVE
        specifies the default form of modulation used by the GatingMechanism's `GatingSignals <GatingSignal>`,
        unless they are `individually specified <GatingSignal_Specification>`.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that can be used to specify the parameters
        for the Mechanism, parameters for its function, and/or a custom function and its parameters. Values
        specified for parameters in the dictionary override any assigned to those parameters in arguments of the
        constructor.

    name : str : default see `name <GatingMechanism.name>`
        specifies the name of the GatingMechanism.

    prefs : PreferenceSet or specification dict : default Mechanism.classPreferences
        specifies the `PreferenceSet` for the GatingMechanism; see `prefs <GatingMechanism.prefs>` for details.


    Attributes
    ----------

    variable : value, list or ndarray
        used as the input to the GatingMechanism's `function <GatingMechanism.function>`.  Its format is determined
        by the **default_gating_allocation** or **size** argument of the GatingMechanism's constructor (see above),
        and is the same format as its `gating_allocation <GatingMechanis.gating_allocation>` (unless a custom
        `function <GatingMechanism.function>` has been assigned).

    monitor_for_gating : List[OutputPort]
        each item is an `OutputPort` monitored by the GatingMechanism or its `objective_mechanism
        <ControlMechanism.objective_mechanism>` if that is specified (see `ControlMechanism_ObjectiveMechanism`);
        in the latter case, the list returned is ObjectiveMechanism's `monitor <ObjectiveMechanism.monitor>` attribute.

    function : TransferFunction
        determines the function used to transform the GatingMechanism's `variable <GatingMechanism.variable>`
        to a `gating_allocation`;  the default is an identity function that simply assigns
        `variable <GatingMechanism.variable>` as the `gating_allocation <GatingMechanism.gating_allocation>`.

    default_allocation : number, list or 1d array
        determines the default_allocation of any `gating_signals <GatingMechanism.gating.signals>` for
        which the **default_allocation** was not specified in its constructor;  if it is None (not specified)
        then the GatingSignal's parameters.allocation.default_value is used. See documentation for
        **default_allocation** argument of GatingSignal constructor for additional details.

    gating_allocation : 2d array
        each item is the value assigned as the `allocation <GatingSignal.allocation>` for the corresponding
        `GatingSignal` listed in the `gating_signals <GatingMechanism.gating_signals>` attribute;  the
        gating_allocation is the same as the GatingMechanism's `value <Mechanism_Base.value>` attribute).

    gating_signals : ContentAddressableList[GatingSignal]
        list of `GatingSignals <GatingSignals>` for the GatingMechanism, each of which sends
        `GatingProjection(s) <GatingProjection>` to the `InputPort(s) <InputPort>` and/or `OutputPorts <OutputPort>`
        that it gates; same as GatingMechanism `output_ports <Mechanism_Base.output_ports>` attribute.

    gating_projections : List[GatingProjection]
        list of all of the `GatingProjections <GatingProjection>` assigned to the GatingMechanism's
        `GatingSignals <GatingSignal>` (i.e., listed in its `gating_signals <GatingMechanism.gating_signals>` attribute.

    value : scalar or 1d np.array of ints
        the result of the GatingMechanism's `function <GatingProjection.funtion>`;
        each item is the value assigned to the corresponding GatingSignal listed in `gating_signals
        <GatingMechanism.gating_signals>`, and used by each GatingSignal to generate the `gating_signal
        <GatingSignal.gating_signal>` assigned to its `GatingProjections <GatingProjection>`; same as the
        GatingMechanism's `gating_allocation <GatingMechanism.gating_allocation>` attribute. Default is a
        single item used by all of the `gating_signals <GatingMechanism.gating_signals>`.

    gating_allocation : scalar or 1d np.array of ints
        the result of the GatingMechanism's `function <Projection_Base.function>`;
        each item is the value assigned to the corresponding `GatingSignal` listed in `gating_signals
        <GatingMechanism.gating_signals>`, and used by each GatingSignal to generate the `gating_signal
        <GatingSignal.gating_signal>` assigned to its `GatingProjections <GatingProjection>`; same as the
        GatingMechanism's `value <GatingMechanism.value>` attribute. Default is a single item used by all of the
        `gating_signals <GatingMechanism.gating_signals>`.

    modulation : str
        the default form of modulation used by the GatingMechanism's `GatingSignals <GatingSignal>`,
        unless they are `individually specified <GatingSignal_Specification>`.

    name : str
        the name of the GatingMechanism; if it is not specified in the **name** argument of the constructor, a default
        is assigned by MechanismRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the GatingMechanism; if it is not specified in the **prefs** argument of the
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentType = "GatingMechanism"

    initMethod = INIT_EXECUTE_METHOD_ONLY

    outputPortTypes = GatingSignal
    portListAttr = ControlMechanism.portListAttr.copy()
    portListAttr.update({GatingSignal:GATING_SIGNALS})


    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'GatingMechanismClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    # Override gating_allocatdion and suppress control_allocation
    class Parameters(ControlMechanism.Parameters):
        """
            Attributes
            ----------

                value
                    see `value <GatingMechanism.value>`

                    :default value: numpy.array([0.5])
                    :type: ``numpy.ndarray``

                gating_allocation
                    see `gating_allocation <GatingMechanism.gating_allocation>`

                    :default value: numpy.array([0.5])
                    :type: ``numpy.ndarray``
                    :read only: True
        """
        # This must be a list, as there may be more than one (e.g., one per control_signal)
        value = Parameter(np.array([defaultGatingAllocation]), aliases='control_allocation', pnl_internal=True)
        gating_allocation = Parameter(
            np.array([defaultGatingAllocation]),
            getter=_gating_allocation_getter,
            setter=_gating_allocation_setter,
            read_only=True,
            pnl_internal=True
        )

    @tc.typecheck
    def __init__(self,
                 default_gating_allocation=None,
                 size=None,
                 monitor_for_gating=None,
                 function=None,
                 default_allocation:tc.optional(tc.any(int, float, list, np.ndarray))=None,
                 gate:tc.optional(tc.optional(list)) = None,
                 modulation:tc.optional(str)=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):

        gate = convert_to_list(gate) or []

        if kwargs:
            # For backward compatibility:
            if GATING_SIGNALS in kwargs:
                args = kwargs.pop(GATING_SIGNALS)
                if args:
                    gate.extend(convert_to_list(args))
            if MONITOR_FOR_CONTROL in kwargs:
                args = kwargs.pop(MONITOR_FOR_CONTROL)
                if args:
                    monitor_for_gating.extend(convert_to_list(args))

        super().__init__(default_variable=default_gating_allocation,
                         size=size,
                         monitor_for_control=monitor_for_gating,
                         function=function,
                         default_allocation=default_allocation,
                         control=gate,
                         modulation=modulation,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

    def _register_control_signal_type(self, context=None):
        from psyneulink.core.globals.registry import register_category
        from psyneulink.core.components.ports.port import Port_Base

        # Create registry for GatingSignals (to manage names)
        register_category(entry=GatingSignal,
                          base_class=Port_Base,
                          registry=self._portRegistry,
                          )

    def _instantiate_control_signal_type(self, gating_signal_spec, context):
        """Instantiate actual ControlSignal, or subclass if overridden"""
        from psyneulink.core.components.ports.port import _instantiate_port
        from psyneulink.core.components.projections.projection import ProjectionError

        allocation_parameter_default = self.parameters.gating_allocation.default_value
        gating_signal = _instantiate_port(port_type=GatingSignal,
                                               owner=self,
                                               variable=self.default_allocation           # User specified value
                                                        or allocation_parameter_default,  # Parameter default
                                               reference_value=allocation_parameter_default,
                                               modulation=self.defaults.modulation,
                                               port_spec=gating_signal_spec,
                                               context=context)
        if not type(gating_signal) in convert_to_list(self.outputPortTypes):
            raise ProjectionError(f'{type(gating_signal)} inappropriate for {self.name}')
        return gating_signal

    def _check_for_duplicates(self, control_signal, control_signals, context):
        """Override ControlMechanism to check in self.gating_signals rather than self.control_signals"""
        super()._check_for_duplicates(control_signal, self.gating_signals, context)

    # Overrided gating_signals
    @property
    def gating_signals(self):
        try:
            return ContentAddressableList(component_type=GatingSignal,
                                          list=[port for port in self.output_ports
                                                if isinstance(port, GatingSignal)])
        except:
            return None

    @gating_signals.setter
    def gating_signals(self, value):
        self._control_signals = value
